# ruff: noqa: S101
import numpy as np
from faster_whisper.transcribe import (
    Segment as WhisperSegment,
)
from faster_whisper.transcribe import (
    Word as WhisperWord,
)

from aana.core.models.asr import (
    AsrSegment,
    AsrSegments,
    AsrTranscription,
    AsrTranscriptionInfo,
    AsrWord,
)
from aana.core.models.time import TimeInterval


def test_asr_segment_from_whisper():
    """Test function for the AsrSegment class's from_whisper method."""
    whisper_segment = WhisperSegment(
        id=0,
        seek=0,
        tokens=[],
        temperature=0.0,
        compression_ratio=0.0,
        start=0.0,
        end=1.0,
        avg_logprob=-0.5,
        no_speech_prob=0.1,
        words=[],
        text="hello world",
    )

    asr_segment = AsrSegment.from_whisper(whisper_segment)

    assert asr_segment.text == "hello world"
    assert asr_segment.time_interval == TimeInterval(
        start=whisper_segment.start, end=whisper_segment.end
    )
    assert asr_segment.confidence == np.exp(whisper_segment.avg_logprob)
    assert asr_segment.no_speech_confidence == whisper_segment.no_speech_prob
    assert asr_segment.words == []

    word = WhisperWord(
        word="hello",
        start=0.0,
        end=0.5,
        probability=0.5,
    )
    whisper_segment = WhisperSegment(
        id=0,
        seek=0,
        tokens=[],
        temperature=0.0,
        compression_ratio=0.0,
        start=0.0,
        end=1.0,
        avg_logprob=-0.5,
        no_speech_prob=0.1,
        words=[word],
        text="hello world",
    )

    asr_segment = AsrSegment.from_whisper(whisper_segment)
    assert asr_segment.words == [AsrWord.from_whisper(word)]


def test_asr_word_from_whisper():
    """Test function for the AsrWord class's from_whisper method."""
    word = WhisperWord(
        word="hello",
        start=0.0,
        end=0.5,
        probability=0.5,
    )

    asr_word = AsrWord.from_whisper(word)

    assert asr_word.word == "hello"
    assert asr_word.time_interval == TimeInterval(start=word.start, end=word.end)
    assert asr_word.alignment_confidence == word.probability


def test_sum_asr_transcription_info():
    """Test function for the sum method of the AsrTranscriptionInfo class."""
    info_1 = AsrTranscriptionInfo(language="en", language_confidence=0.9)
    info_2 = AsrTranscriptionInfo(language="de", language_confidence=0.8)

    info_sum = sum([info_1, info_2], AsrTranscriptionInfo())

    assert info_sum.language == "en"
    assert info_sum.language_confidence == 0.9

    info_1 = AsrTranscriptionInfo(language="en", language_confidence=0.9)
    info_2 = AsrTranscriptionInfo(language="en", language_confidence=0.5)

    info_sum = sum([info_1, info_2], AsrTranscriptionInfo())

    assert info_sum.language == "en"
    assert info_sum.language_confidence == 0.7


def test_sum_asr_segments():
    """Test function for the sum method of the AsrSegments class."""
    whisper_segment = WhisperSegment(
        id=0,
        seek=0,
        tokens=[],
        temperature=0.0,
        compression_ratio=0.0,
        start=0.0,
        end=1.0,
        avg_logprob=-0.5,
        no_speech_prob=0.1,
        words=[],
        text="hello world",
    )

    asr_segment_1 = AsrSegment.from_whisper(whisper_segment)
    asr_segment_2 = AsrSegment.from_whisper(whisper_segment)

    segment_1 = AsrSegments([asr_segment_1] * 3)
    segment_2 = AsrSegments([asr_segment_2] * 3)

    segments = sum([segment_1, segment_2], AsrSegments())

    assert len(segments) == 6
    assert segments == [asr_segment_1] * 3 + [asr_segment_2] * 3
    assert segments[:3] == segment_1
    assert segments[3:] == segment_2


def test_sum_asr_transcription():
    """Test function for the sum method of the AsrTranscription class."""
    transcription1 = AsrTranscription(text="Hello world")
    transcription2 = AsrTranscription(text="Another transcription")

    transcription = sum([transcription1, transcription2], AsrTranscription())

    assert transcription.text == "Hello world\nAnother transcription"


def test_asr_segment():
    """Test function for the AsrSegment class."""
    segment = AsrSegment(
        text="Hello world",
        time_interval=TimeInterval(start=0.0, end=1.0),
        confidence=0.5,
        no_speech_confidence=0.1,
    )
    assert segment.text == "Hello world"
    assert segment.time_interval == TimeInterval(start=0.0, end=1.0)
    assert segment.confidence == 0.5
    assert segment.no_speech_confidence == 0.1

    # Test that the segment can be created without confidence and no_speech_confidence
    segment = AsrSegment(
        text="Hello world", time_interval=TimeInterval(start=0.0, end=1.0)
    )

    assert segment.text == "Hello world"
    assert segment.time_interval == TimeInterval(start=0.0, end=1.0)



from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from aana.core.models.media import MediaId
from aana.storage.models.media import MediaEntity


class VideoEntity(MediaEntity):
    """Base ORM class for videos.

    Attributes:
        id (MediaId): Unique identifier for the video.
        path (str): Path to the video file.
        url (str): URL to the video file.
        title (str): Title of the video.
        description (str): Description of the video.
    """

    __tablename__ = "video"

    id: Mapped[MediaId] = mapped_column(ForeignKey("media.id"), primary_key=True)
    path: Mapped[str] = mapped_column(comment="Path", nullable=True)
    url: Mapped[str] = mapped_column(comment="URL", nullable=True)
    title: Mapped[str] = mapped_column(comment="Title", nullable=True)
    description: Mapped[str] = mapped_column(comment="Description", nullable=True)

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "video",
    }



from enum import Enum

import torch

__all__ = ["Dtype"]


class Dtype(str, Enum):
    """Data types.

    Possible values are "auto", "float32", "float16", "bfloat16" and "int8".

    Attributes:
        AUTO (str): auto
        FLOAT32 (str): float32
        FLOAT16 (str): float16
        BFLOAT16 (str): bfloat16
        INT8 (str): int8
    """

    AUTO = "auto"
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    BFLOAT16 = "bfloat16"
    INT8 = "int8"

    def to_torch(self) -> torch.dtype | str:
        """Convert the instance's dtype to a torch dtype.

        Returns:
            Union[torch.dtype, str]: the torch dtype or "auto"

        Raises:
            ValueError: if the dtype is unknown
        """
        match self.value:
            case self.AUTO:
                return "auto"
            case self.FLOAT32:
                return torch.float32
            case self.FLOAT16:
                return torch.float16
            case self.BFLOAT16:
                return torch.bfloat16
            case self.INT8:
                return torch.int8
            case _:
                raise ValueError(f"Unknown dtype: {self}")  # noqa: TRY003



import hashlib  # noqa: I001
from dataclasses import dataclass
from pathlib import Path
import torch, decord  # noqa: F401  # See https://github.com/dmlc/decord/issues/263
from decord import DECORDError
from typing import Annotated
from aana.configs.settings import settings
from aana.exceptions.io import VideoReadingException
from aana.core.models.media import Media
import uuid

from pydantic import (
    AfterValidator,
    AnyUrl,
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
)
from pydantic_core import InitErrorDetails
from typing_extensions import Self

from aana.core.models.base import BaseListModel
from aana.core.models.media import MediaId

__all__ = ["VideoMetadata", "VideoParams"]


@dataclass
class Video(Media):
    """A class representing a video.

    At least one of `path`, `url`, or `content` must be provided.
    If `save_on_disk` is True, the video will be saved on disk automatically.

    Attributes:
        path (Path): the path to the video file
        url (str): the URL of the video
        content (bytes): the content of the video in bytes
        media_id (MediaId): the ID of the video. If not provided, it will be generated automatically.
        title (str): the title of the video
        description (str): the description of the video
        media_dir (Path): the directory to save the video in
    """

    title: str = ""
    description: str = ""
    media_dir: Path | None = settings.video_dir

    def _validate(self):
        """Validate the video.

        Raises:
            ValueError: if none of 'path', 'url', or 'content' is provided
            VideoReadingException: if the video is not valid
        """
        # validate the parent class
        super()._validate()

        # check that at least one of 'path', 'url' or 'content' is provided
        if not any(
            [
                self.path is not None,
                self.url is not None,
                self.content is not None,
            ]
        ):
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url' or 'content' must be provided."
            )

        # check that the video is valid
        if self.path and not self.is_video():
            raise VideoReadingException(video=self)

    def is_video(self) -> bool:
        """Checks if it's a valid video."""
        if not self.path:
            return False

        try:
            decord.VideoReader(str(self.path))
        except DECORDError:
            try:
                decord.AudioReader(str(self.path))
            except DECORDError:
                return False
        return True

    def _save_from_url(self, file_path):
        """Save the media from the URL.

        Args:
            file_path (Path): the path to save the media to

        Raises:
            DownloadError: if the media can't be downloaded
            VideoReadingException: if the media is not a valid video
        """
        super()._save_from_url(file_path)
        # check that the file is a video
        if not self.is_video():
            raise VideoReadingException(video=self)

    def __repr__(self) -> str:
        """Get the representation of the video.

        Use md5 hash for the content of the video if it is available.

        Returns:
            str: the representation of the video
        """
        content_hash = (
            hashlib.md5(self.content, usedforsecurity=False).hexdigest()
            if self.content
            else None
        )
        return (
            f"Video(path={self.path}, "
            f"url={self.url}, "
            f"content={content_hash}, "
            f"media_id={self.media_id}, "
            f"title={self.title})"
        )


class VideoMetadata(BaseModel):
    """Metadata of a video.

    Attributes:
        title (str): the title of the video
        description (str): the description of the video
        duration (float): the duration of the video in seconds
    """

    title: str = Field(None, description="The title of the video.")
    description: str = Field(None, description="The description of the video.")
    duration: float | None = Field(
        None, description="The duration of the video in seconds."
    )
    model_config = ConfigDict(
        json_schema_extra={
            "description": "Metadata of a video.",
        },
        extra="forbid",
    )


class VideoParams(BaseModel):
    """A pydantic model for video parameters.

    Attributes:
        extract_fps (float): the number of frames to extract per second
        fast_mode_enabled (bool): whether to use fast mode (keyframes only)
    """

    extract_fps: float = Field(
        default=3.0,
        gt=0.0,
        description=(
            "The number of frames to extract per second. "
            "Can be smaller than 1. For example, 0.5 means 1 frame every 2 seconds."
        ),
    )
    fast_mode_enabled: bool = Field(
        default=True,
        description=(
            "Whether to use fast mode (keyframes only). "
            "extract_fps will be ignored if this is set to True."
        ),
    )
    model_config = ConfigDict(
        json_schema_extra={"description": "Video parameters."}, validate_assignment=True
    )


class VideoInput(BaseModel):
    """A video input.

    Exactly one of 'path', 'url', or 'content' must be provided.

    If 'content' is set to 'file',
    the video will be loaded from the files uploaded to the endpoint.

    Attributes:
        media_id (MediaId): the ID of the video. If not provided, it will be generated automatically.
        path (str): the file path of the video
        url (AnyUrl): the URL of the video (supports YouTube videos)
        content (bytes): the content of the video in bytes
    """

    path: str | None = Field(None, description="The file path of the video.")
    url: Annotated[
        AnyUrl | None,
        AfterValidator(lambda x: str(x) if x else None),
        Field(None, description="The URL of the video (supports YouTube videos)."),
    ]
    content: bytes | None = Field(
        None,
        description=(
            "The content of the video in bytes. "
            "Set this field to 'file' to upload files to the endpoint."
        ),
    )
    media_id: MediaId = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description="The ID of the video. If not provided, it will be generated automatically.",
    )

    @model_validator(mode="after")
    def check_only_one_field(self) -> Self:
        """Check that exactly one of 'path', 'url', or 'content' is provided.

        Raises:
            ValueError: if not exactly one of 'path', 'url', or 'content' is provided

        Returns:
            Self: the instance
        """
        count = sum(value is not None for value in [self.path, self.url, self.content])
        if count != 1:
            raise ValueError(  # noqa: TRY003
                "Exactly one of 'path', 'url', or 'content' must be provided."
            )
        return self

    def set_file(self, file: bytes):
        """Sets the file.

        If 'content' is set to 'file',
        the video will be loaded from the file uploaded to the endpoint.

        set_file() should be called after the files are uploaded to the endpoint.

        Args:
            file (bytes): the file uploaded to the endpoint

        Raises:
            ValueError: if the content isn't set to 'file'
        """
        if self.content == b"file":
            self.content = file
        else:
            raise ValueError("The content of the video must be 'file' to set files.")  # noqa: TRY003

    def set_files(self, files: list[bytes]):
        """Set the files for the video.

        Args:
            files (List[bytes]): the files uploaded to the endpoint

        Raises:
            ValidationError: if the number of files isn't 1
        """
        if len(files) != 1:
            raise ValidationError.from_exception_data(
                title=self.__class__.__name__,
                line_errors=[
                    InitErrorDetails(
                        loc=("video",),
                        type="value_error",
                        ctx={
                            "error": ValueError(
                                "The number of videos and files must be the same."
                            )
                        },
                        input=None,
                    )
                ],
            )
        self.set_file(files[0])

    def convert_input_to_object(self) -> Video:
        """Convert the video input to a video object.

        Returns:
            Video: the video object corresponding to the video input
        """
        if self.content == b"file":
            raise ValueError(  # noqa: TRY003
                "The content of the video isn't set. Please upload files and call set_files()."
            )
        return Video(
            path=Path(self.path) if self.path is not None else None,
            url=self.url,
            content=self.content,
            media_id=self.media_id,
        )

    model_config = ConfigDict(
        json_schema_extra={
            "description": (
                "A video. \n"
                "Exactly one of 'path', 'url', or 'content' must be provided. \n"
                "If 'path' is provided, the video will be loaded from the path. \n"
                "If 'url' is provided, the video will be downloaded from the url. \n"
                "The 'content' will be loaded automatically "
                "if files are uploaded to the endpoint (should be set to 'file' for that)."
            )
        },
        validate_assignment=True,
        file_upload=True,
        file_upload_description="Upload video file.",
    )


class VideoInputList(BaseListModel):
    """A pydantic model for a list of video inputs.

    Only used for the requests, DO NOT use it for anything else.

    Convert it to a list of video objects with convert_input_to_object().
    """

    root: list[VideoInput]

    @model_validator(mode="after")
    def check_non_empty(self) -> Self:
        """Check that the list of videos isn't empty.

        Raises:
            ValueError: if the list of videos is empty

        Returns:
            Self: the instance
        """
        if len(self.root) == 0:
            raise ValueError("The list of videos must not be empty.")  # noqa: TRY003
        return self

    def set_files(self, files: list[bytes]):
        """Set the files for the videos.

        Args:
            files (List[bytes]): the files uploaded to the endpoint

        Raises:
            ValidationError: if the number of videos and files aren't the same
        """
        if len(self.root) != len(files):
            raise ValidationError.from_exception_data(
                title=self.__class__.__name__,
                line_errors=[
                    InitErrorDetails(
                        loc=("videos",),
                        type="value_error",
                        ctx={
                            "error": ValueError(
                                "The number of videos and files must be the same."
                            )
                        },
                        input=None,
                    )
                ],
            )
        for video, file in zip(self.root, files, strict=False):
            video.set_file(file)

    def convert_input_to_object(self) -> list[VideoInput]:
        """Convert the VideoInputList to a list of video inputs.

        Returns:
            List[VideoInput]: the list of video inputs
        """
        return self.root

    model_config = ConfigDict(
        json_schema_extra={
            "description": (
                "A list of videos. \n"
                "Exactly one of 'path', 'url', or 'content' must be provided for each video. \n"
                "If 'path' is provided, the video will be loaded from the path. \n"
                "If 'url' is provided, the video will be downloaded from the url. \n"
                "The 'content' will be loaded automatically "
                "if files are uploaded to the endpoint (should be set to 'file' for that)."
            )
        },
        file_upload=True,
        file_upload_description="Upload video files.",
    )



# ruff: noqa: S101

import pytest

from aana.core.models.asr import AsrSegment, AsrTranscription, AsrTranscriptionInfo
from aana.core.models.time import TimeInterval
from aana.exceptions.db import NotFoundException
from aana.storage.models.transcript import TranscriptEntity
from aana.storage.repository.transcript import TranscriptRepository

transcript_entity = TranscriptEntity.from_asr_output(
    model_name="whisper",
    transcription=AsrTranscription(text="This is a transcript"),
    segments=[],
    info=AsrTranscriptionInfo(),
)


@pytest.fixture(scope="function")
def dummy_transcript():
    """Creates a dummy transcript for testing."""
    transcript = AsrTranscription(text="This is a transcript")
    segments = [
        AsrSegment(text="This is a segment", time_interval=TimeInterval(start=0, end=1))
    ]
    info = AsrTranscriptionInfo(language="en", language_confidence=0.9)
    return transcript, segments, info


def test_save_transcript(db_session, dummy_transcript):
    """Tests saving a transcript."""
    transcript, segments, info = dummy_transcript
    model_name = "whisper"

    transcript_repo = TranscriptRepository(db_session)
    transcript_entity = transcript_repo.save(
        model_name=model_name,
        transcription_info=info,
        transcription=transcript,
        segments=segments,
    )

    transcript_id = transcript_entity.id

    transcript_entity = transcript_repo.read(transcript_id)
    assert transcript_entity
    assert transcript_entity.id == transcript_id
    assert transcript_entity.model == model_name
    assert transcript_entity.transcript == transcript.text
    assert len(transcript_entity.segments) == len(segments)
    assert transcript_entity.language == info.language
    assert transcript_entity.language_confidence == info.language_confidence

    transcript_repo.delete(transcript_id)
    with pytest.raises(NotFoundException):
        transcript_repo.read(transcript_id)



# ruff: noqa: S101
from collections.abc import Callable

import pytest
from typing_extensions import override

from aana.api.event_handlers.event_handler import EventHandler
from aana.api.event_handlers.event_manager import EventManager
from aana.exceptions.runtime import (
    HandlerNotRegisteredException,
)


class CallbackHandler(EventHandler):
    """A test event handler that just invokes a callback function."""

    def __init__(self, callback: Callable):
        """Constructor."""
        super().__init__()
        self.callback = callback

    @override
    def handle(self, event_name: str, *args, **kwargs):
        self.callback(event_name, *args, **kwargs)


def test_event_dispatch():
    """Tests that event dispatch works correctly."""
    event_manager = EventManager()
    expected_event_name = "foo"
    expected_args = (1, 2, 3, 4, 5)
    expected_kwargs = {"a": "A", "b": "B"}

    def callback(event_name, *args, **kwargs):
        assert event_name == expected_event_name
        assert args == expected_args
        assert kwargs == expected_kwargs

    event_manager.register_handler_for_events(CallbackHandler(callback), ["foo"])

    event_manager.handle("foo", 1, 2, 3, 4, 5, a="A", b="B")


def test_remove_all_raises():
    """Tests that removing handler not added from all events raises an error."""
    event_manager = EventManager()
    handler = CallbackHandler(lambda _, *_args, **_kwargs: None)
    with pytest.raises(HandlerNotRegisteredException):
        event_manager.deregister_handler_from_all_events(handler)


def test_remove_works():
    """Tests that removing a handler works."""
    event_manager = EventManager()
    handler = CallbackHandler(lambda _, *_args, **_kwargs: None)
    event_manager.register_handler_for_events(handler, ["foo"])
    event_manager.deregister_handler_from_event(handler, "foo")
    assert len(event_manager._handlers["foo"]) == 0


def test_remove_all_works():
    """Tests that removing all handlers works."""
    event_manager = EventManager()
    handler = CallbackHandler(lambda _, *_args, **_kwargs: None)
    event_manager.register_handler_for_events(handler, ["foo"])
    event_manager.deregister_handler_from_all_events(handler)
    assert len(event_manager._handlers["foo"]) == 0



from pathlib import Path

from pydantic import BaseModel, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from aana.configs.db import DbSettings
from aana.core.models.base import pydantic_protected_fields


class TestSettings(BaseModel):
    """A pydantic model for test settings.

    Attributes:
        test_mode (bool): Flag indicating if the SDK is in test mode.
        save_expected_output (bool): Flag indicating if the expected output should be saved (to create test cases).
    """

    test_mode: bool = False
    save_expected_output: bool = False


class TaskQueueSettings(BaseModel):
    """A pydantic model for task queue settings.

    Attributes:
        enabled (bool): Flag indicating if the task queue is enabled.
        num_workers (int): The number of workers in the task queue.
        execution_timeout (int): The maximum execution time for a task in seconds.
            After this time, if the task is still running,
            it will be considered as stuck and will be reassign to another worker.
        heartbeat_timeout (int): The maximum time between heartbeats in seconds.
        max_retries (int): The maximum number of retries for a task.
    """

    enabled: bool = True
    num_workers: int = 4
    execution_timeout: int = 600
    heartbeat_timeout: int = 60
    max_retries: int = 3


class Settings(BaseSettings):
    """A pydantic model for SDK settings.

    Attributes:
        tmp_data_dir (Path): The temporary data directory.
        image_dir (Path): The temporary image directory.
        video_dir (Path): The temporary video directory.
        audio_dir (Path): The temporary audio directory.
        model_dir (Path): The temporary model directory.
        num_workers (int): The number of web workers.
        task_queue (TaskQueueSettings): The task queue settings.
        db_config (DbSettings): The database configuration.
        test (TestSettings): The test settings.
    """

    tmp_data_dir: Path = Path("/tmp/aana_data")  # noqa: S108
    image_dir: Path = tmp_data_dir / "images"
    video_dir: Path = tmp_data_dir / "videos"
    audio_dir: Path = tmp_data_dir / "audios"
    model_dir: Path = tmp_data_dir / "models"
    num_workers: int = 2

    task_queue: TaskQueueSettings = TaskQueueSettings()

    db_config: DbSettings = DbSettings()

    test: TestSettings = TestSettings()

    @field_validator("tmp_data_dir", mode="after")
    def create_tmp_data_dir(cls, path: Path) -> Path:
        """Create the tmp_data_dir if it doesn't exist."""
        path.mkdir(parents=True, exist_ok=True)
        return path

    model_config = SettingsConfigDict(
        protected_namespaces=("settings", *pydantic_protected_fields),
        env_nested_delimiter="__",
        env_ignore_empty=True,
    )


settings = Settings()



# ruff: noqa: S101
from importlib import resources

import pytest
from transformers import BitsAndBytesConfig

from aana.core.models.image import Image
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.hf_pipeline_deployment import (
    HfPipelineConfig,
    HfPipelineDeployment,
)
from aana.tests.utils import verify_deployment_results

deployments = [
    (
        "hf_pipeline_blip2_deployment",
        HfPipelineDeployment.options(
            num_replicas=1,
            ray_actor_options={"num_gpus": 1},
            user_config=HfPipelineConfig(
                model_id="Salesforce/blip2-opt-2.7b",
                model_kwargs={
                    "quantization_config": BitsAndBytesConfig(
                        load_in_8bit=False, load_in_4bit=True
                    ),
                },
            ).model_dump(mode="json"),
        ),
    )
]


@pytest.mark.parametrize("setup_deployment", deployments, indirect=True)
class TestHFPipelineDeployment:
    """Test HuggingFace Pipeline deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("image_name", ["Starry_Night.jpeg"])
    async def test_call(self, setup_deployment, image_name):
        """Test call method."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "hf_pipeline"
            / f"{deployment_name}_{image_name}.json"
        )
        path = resources.files("aana.tests.files.images") / image_name
        image = Image(path=path, save_on_disk=False, media_id=image_name)

        output = await handle.call(images=image)
        verify_deployment_results(expected_output_path, output)

        output = await handle.call(image)
        verify_deployment_results(expected_output_path, output)

        output = await handle.call(images=[str(path)])
        verify_deployment_results(expected_output_path, [output])

        output = await handle.call(images=[image])
        verify_deployment_results(expected_output_path, [output])

        output = await handle.call([image])
        verify_deployment_results(expected_output_path, [output])



from typing import Any

from pydantic import BaseModel, Field

from aana.storage.models.task import Status as TaskStatus


class TaskInfo(BaseModel):
    """Task information.

    Attributes:
        id (str): The task ID.
        status (TaskStatus): The task status.
        result (Any): The task result.
    """

    id: str = Field(..., description="The task ID.")
    status: TaskStatus = Field(..., description="The task status.")
    result: Any = Field(None, description="The task result.")


class TaskId(BaseModel):
    """Task ID.

    Attributes:
        id (str): The task ID.
    """

    task_id: str = Field(..., description="The task ID.")



# ruff: noqa: S101
import pytest

from aana.configs.db import DbSettings, PostgreSQLConfig, SQLiteConfig


@pytest.fixture
def pg_settings():
    """Fixture for working PostgreSQL settings."""
    return DbSettings(
        datastore_type="postgresql",
        datastore_config=PostgreSQLConfig(
            host="0.0.0.0",  # noqa: S104
            port="5432",
            database="postgres",
            user="postgres",
            password="bogus",  # noqa: S106
        ),
    )


@pytest.fixture
def sqlite_settings():
    """Fixture for working sqlite settings."""
    return DbSettings(
        datastore_type="sqlite",
        datastore_config=SQLiteConfig(path="/tmp/deleteme.sqlite"),  # noqa: S108
    )


def test_get_engine_idempotent(pg_settings, sqlite_settings):
    """Tests that get_engine returns the same engine on subsequent calls."""
    for db_settings in (pg_settings, sqlite_settings):
        e1 = db_settings.get_engine()
        e2 = db_settings.get_engine()
        assert e1 is e2


def test_pg_datastore_config(pg_settings):
    """Tests datastore config for postgres."""
    engine = pg_settings.get_engine()

    assert engine.name == "postgresql"
    assert str(engine.url) == "postgresql+psycopg://postgres:***@0.0.0.0:5432/postgres"


def test_sqlite_datastore_config(sqlite_settings):
    """Tests datastore config for SQLite."""
    engine = sqlite_settings.get_engine()

    assert engine.name == "sqlite"
    assert str(engine.url) == f"sqlite:///{sqlite_settings.datastore_config['path']}"


def test_nonexistent_datastore_config():
    """Tests that datastore config errors on unsupported DB types."""
    db_settings = DbSettings(
        **{
            "datastore_type": "oracle",
            "datastore_config": {
                "host": "0.0.0.0",  # noqa: S104
                "port": "5432",
                "database": "oracle",
                "user": "oracle",
                "password": "bogus",
            },
        }
    )
    with pytest.raises(ValueError):
        _ = db_settings.get_engine()


def test_invalid_datastore_config(pg_settings, sqlite_settings):
    """Tests that a datastore with the wrong config raises errors."""
    tmp = pg_settings.datastore_config
    pg_settings.datastore_config = sqlite_settings.datastore_config
    sqlite_settings.datastore_config = tmp

    with pytest.raises(KeyError):
        _ = sqlite_settings.get_engine()
    with pytest.raises(KeyError):
        _ = pg_settings.get_engine()



import inspect
from collections.abc import Callable

import ray

__all__ = ["run_remote"]


def run_remote(func: Callable) -> Callable:
    """Wrap a function to run it remotely on Ray.

    Args:
        func (Callable): the function to wrap

    Returns:
        Callable: the wrapped function
    """

    async def generator_wrapper(*args, **kwargs):
        async for item in ray.remote(func).remote(*args, **kwargs):
            yield await item

    if inspect.isgeneratorfunction(func):
        return generator_wrapper
    else:
        return ray.remote(func).remote



# ruff: noqa: S101
from importlib import resources

import pytest

from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.image import Image
from aana.core.models.image_chat import ImageChatDialog
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.vllm_deployment import VLLMConfig, VLLMDeployment
from aana.exceptions.runtime import PromptTooLongException
from aana.tests.utils import verify_deployment_results
from aana.utils.core import get_object_hash

deployments = [
    (
        (
            "phi-3.5-vision-instruct_vllm_deployment",
            VLLMDeployment.options(
                num_replicas=1,
                ray_actor_options={"num_gpus": 1.0},
                user_config=VLLMConfig(
                    model_id="microsoft/Phi-3.5-vision-instruct",
                    dtype=Dtype.FLOAT16,
                    gpu_memory_reserved=12000,
                    enforce_eager=True,
                    default_sampling_params=SamplingParams(
                        temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
                    ),
                    max_model_len=2048,
                    engine_args=dict(
                        trust_remote_code=True,
                        max_num_seqs=32,
                        limit_mm_per_prompt={"image": 3},
                    ),
                ).model_dump(mode="json"),
            ),
        ),
        "<s><|user|>\n{query}<|end|>\n<|assistant|>\n",
    ),
    # disabled because it requires unrealeased version of transformers
    # (
    #     (
    #         "qwen2_vl_7b_instruct_vllm_deployment",
    #         VLLMDeployment.options(
    #             num_replicas=1,
    #             ray_actor_options={"num_gpus": 1.0},
    #             user_config=VLLMConfig(
    #                 model_id="Qwen/Qwen2-VL-7B-Instruct",
    #                 gpu_memory_reserved=40000,
    #                 enforce_eager=True,
    #                 default_sampling_params=SamplingParams(
    #                     temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
    #                 ),
    #                 max_model_len=4096,
    #                 engine_args=dict(
    #                     limit_mm_per_prompt={"image": 3},
    #                 ),
    #             ).model_dump(mode="json"),
    #         ),
    #     ),
    #     "<|im_start|>system\nYou are a helpful assistant.<|im_end|>\n<|im_start|>user\n{query}<|im_end|>\n<|im_start|>assistant\n",
    # ),
    (
        (
            "pixtral_12b_2409_vllm_deployment",
            VLLMDeployment.options(
                num_replicas=1,
                ray_actor_options={"num_gpus": 1.0},
                user_config=VLLMConfig(
                    model_id="mistralai/Pixtral-12B-2409",
                    gpu_memory_reserved=40000,
                    enforce_eager=True,
                    default_sampling_params=SamplingParams(
                        temperature=0.0, top_p=1.0, top_k=-1, max_tokens=1024
                    ),
                    max_model_len=4096,
                    engine_args=dict(
                        tokenizer_mode="mistral",
                        limit_mm_per_prompt={"image": 3},
                    ),
                ).model_dump(mode="json"),
            ),
        ),
        "<s>[INST]{query}[/INST] \n",
    ),
]


@pytest.mark.parametrize(
    "setup_deployment, prompt_template", deployments, indirect=["setup_deployment"]
)
class TestImageTextGenerationDeployments:
    """Test image-text generation deployments."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "prompt, image_name",
        [("Who is the painter of the image?", "Starry_Night.jpeg")],
    )
    async def test_image_text_generation_methods(
        self, setup_deployment, prompt_template, prompt, image_name
    ):
        """Test image-text generation methods."""
        deployment_name, handle_name, _ = setup_deployment
        handle = await AanaDeploymentHandle.create(handle_name)

        prompt_hash = get_object_hash(prompt)
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "image_text_generation"
            / f"{deployment_name}_{image_name}_{prompt_hash}.json"
        )

        image = Image(path=resources.files("aana.tests.files.images") / image_name)
        dialog = ImageChatDialog.from_prompt(prompt=prompt, images=[image])

        # test chat method
        output = await handle.chat(dialog=dialog)
        output_message = output["message"]

        assert isinstance(output_message, ChatMessage)
        assert output_message.role == "assistant"
        verify_deployment_results(expected_output_path, output_message.content)

        # test chat_stream method
        stream = handle.chat_stream(dialog=dialog)
        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("query", ["Who is Elon Musk?"])
    async def test_text_generation_methods(
        self, setup_deployment, prompt_template, query
    ):
        """Test text generation methods."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        query_hash = get_object_hash(query)
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "text_generation"
            / f"{deployment_name}_{query_hash}.json"
        )

        prompt = prompt_template.format(query=query)

        # test generate method
        output = await handle.generate(
            prompt=prompt,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        verify_deployment_results(
            expected_output_path, output["text"], allowed_error_rate=0.6
        )

        # test generate_stream method
        stream = handle.generate_stream(
            prompt=prompt,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text, allowed_error_rate=0.6)

        # test generate_batch method
        output = await handle.generate_batch(
            prompts=[prompt, prompt],
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )
        texts = output["texts"]

        for text in texts:
            verify_deployment_results(
                expected_output_path, text, allowed_error_rate=0.6
            )

        # test chat method
        dialog = ChatDialog(
            messages=[
                ChatMessage(
                    role="user",
                    content=query,
                )
            ]
        )
        output = await handle.chat(
            dialog=dialog,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )

        response_message = output["message"]
        assert response_message.role == "assistant"
        text = response_message.content

        verify_deployment_results(expected_output_path, text)

        # test chat_stream method
        stream = handle.chat_stream(
            dialog=dialog,
            sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
        )

        text = ""
        async for chunk in stream:
            text += chunk["text"]

        verify_deployment_results(expected_output_path, text)

        # test generate method with too long prompt
        with pytest.raises(PromptTooLongException):
            output = await handle.generate(
                prompt=prompt * 1000,
                sampling_params=SamplingParams(temperature=0.0, max_tokens=32),
            )



from aana.configs.db import DbSettings, PostgreSQLConfig, SQLiteConfig
from aana.configs.settings import Settings, TaskQueueSettings, TestSettings
from aana.storage.op import DbType

__all__ = [
    "Settings",
    "DbSettings",
    "TaskQueueSettings",
    "TestSettings",
    "DbType",
    "PostgreSQLConfig",
    "SQLiteConfig",
]



# ruff: noqa: S101, NPY002
import io
from importlib import resources
from pathlib import Path

import numpy as np
import pytest
from pydantic import ValidationError

from aana.core.models.image import ImageInput, ImageInputList


@pytest.fixture
def mock_download_file(mocker):
    """Mock download_file."""
    mock = mocker.patch("aana.core.models.image.download_file", autospec=True)
    path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
    content = path.read_bytes()
    mock.return_value = content
    return mock


def test_new_imageinput_success():
    """Test that ImageInput can be created successfully."""
    image_input = ImageInput(path="image.png")
    assert image_input.path == "image.png"

    image_input = ImageInput(url="http://image.png")
    assert image_input.url == "http://image.png/"

    image_input = ImageInput(content=b"file")
    assert image_input.content == b"file"

    image_input = ImageInput(numpy=b"file")
    assert image_input.numpy == b"file"


def test_imageinput_invalid_media_id():
    """Test that ImageInput can't be created if media_id is invalid."""
    with pytest.raises(ValidationError):
        ImageInput(path="image.png", media_id="")


@pytest.mark.parametrize(
    "url",
    [
        "domain",
        "domain.com",
        "http://",
        "www.domain.com",
        "/subdomain",
        "../subdomain",
        "",
    ],
)
def test_imageinput_invalid_url(url):
    """Test that ImageInput can't be created if url is invalid."""
    with pytest.raises(ValidationError):
        ImageInput(url=url)


def test_imageinput_check_only_one_field():
    """Test that exactly one of 'path', 'url', 'content', or 'numpy' is provided."""
    fields = {
        "path": "image.png",
        "url": "http://image.png",
        "content": b"file",
        "numpy": b"file",
    }

    # check all combinations of two fields
    for field1 in fields:
        for field2 in fields:
            if field1 != field2:
                with pytest.raises(ValidationError):
                    ImageInput(**{field1: fields[field1], field2: fields[field2]})

    # check all combinations of three fields
    for field1 in fields:
        for field2 in fields:
            for field3 in fields:
                if field1 != field2 and field1 != field3 and field2 != field3:
                    with pytest.raises(ValidationError):
                        ImageInput(
                            **{
                                field1: fields[field1],
                                field2: fields[field2],
                                field3: fields[field3],
                            }
                        )

    # check all combinations of four fields
    with pytest.raises(ValidationError):
        ImageInput(**fields)

    # check that no fields is also invalid
    with pytest.raises(ValidationError):
        ImageInput()


def test_imageinput_set_file():
    """Test that the file can be set for the image."""
    file_content = b"image data"

    # If 'content' is set to 'file',
    # the image can be set from the file uploaded to the endpoint.
    image_input = ImageInput(content=b"file")
    image_input.set_file(file_content)
    assert image_input.content == file_content

    # If 'numpy' is set to 'file',
    # the image can be set from the file uploaded to the endpoint.
    image_input = ImageInput(numpy=b"file")
    image_input.set_file(file_content)
    assert image_input.numpy == file_content

    # If neither 'content' nor 'numpy' is set to 'file',
    # an error should be raised.
    image_input = ImageInput(path="image.png")
    with pytest.raises(ValueError):
        image_input.set_file(file_content)


def test_imageinput_set_files():
    """Test that the files can be set for the image."""
    files = [b"image data"]

    # If 'content' is set to 'file',
    # the image can be set from the file uploaded to the endpoint.
    image_input = ImageInput(content=b"file")
    image_input.set_files(files)
    assert image_input.content == files[0]

    # If 'numpy' is set to 'file',
    # the image can be set from the file uploaded to the endpoint.
    image_input = ImageInput(numpy=b"file")
    image_input.set_files(files)
    assert image_input.numpy == files[0]

    # If neither 'content' nor 'numpy' is set to 'file',
    # an error should be raised.
    image_input = ImageInput(path="image.png")
    with pytest.raises(ValueError):
        image_input.set_files(files)

    # If the number of images and files aren't the same,
    # an error should be raised.
    files = [b"image data", b"another image data"]
    image_input = ImageInput(content=b"file")
    with pytest.raises(ValidationError):
        image_input.set_files(files)

    files = []
    image_input = ImageInput(content=b"file")
    with pytest.raises(ValidationError):
        image_input.set_files(files)


def test_imageinput_convert_input_to_object(mock_download_file):
    """Test that ImageInput can be converted to Image."""
    path = resources.files("aana.tests.files.images") / "Starry_Night.jpeg"
    image_input = ImageInput(path=str(path))
    try:
        image_object = image_input.convert_input_to_object()
        assert image_object.path == path
    finally:
        image_object.cleanup()

    url = "http://example.com/Starry_Night.jpeg"
    image_input = ImageInput(url=url)
    try:
        image_object = image_input.convert_input_to_object()
        assert image_object.url == url
    finally:
        image_object.cleanup()

    content = Path(path).read_bytes()
    image_input = ImageInput(content=content)
    try:
        image_object = image_input.convert_input_to_object()
        assert image_object.content == content
    finally:
        image_object.cleanup()

    numpy = np.random.rand(100, 100, 3).astype(np.uint8)
    # convert numpy array to bytes
    buffer = io.BytesIO()
    np.save(buffer, numpy)
    numpy_bytes = buffer.getvalue()
    image_input = ImageInput(numpy=numpy_bytes)
    try:
        image_object = image_input.convert_input_to_object()
        assert np.array_equal(image_object.numpy, numpy)
    finally:
        image_object.cleanup()


def test_imageinput_convert_input_to_object_invalid_numpy():
    """Test that ImageInput can't be converted to Image if numpy is invalid."""
    numpy = np.random.rand(100, 100, 3).astype(np.uint8)
    # convert numpy array to bytes
    buffer = io.BytesIO()
    np.save(buffer, numpy)
    numpy_bytes = buffer.getvalue()
    # remove the last byte
    numpy_bytes = numpy_bytes[:-1]
    image_input = ImageInput(numpy=numpy_bytes)
    with pytest.raises(ValueError):
        image_input.convert_input_to_object()


def test_imageinput_convert_input_to_object_numpy_not_set():
    """Test that ImageInput can't be converted to Image if numpy file isn't set with set_file()."""
    image_input = ImageInput(numpy=b"file")
    with pytest.raises(ValueError):
        image_input.convert_input_to_object()


def test_imagelistinput():
    """Test that ImageListInput can be created successfully."""
    images = [
        ImageInput(path="image.png"),
        ImageInput(url="http://image.png"),
        ImageInput(content=b"file"),
        ImageInput(numpy=b"file"),
    ]

    image_list_input = ImageInputList(images)
    assert image_list_input.root == images
    assert len(image_list_input) == len(images)
    assert image_list_input[0] == images[0]
    assert image_list_input[1] == images[1]
    assert image_list_input[2] == images[2]
    assert image_list_input[3] == images[3]


def test_imagelistinput_set_files():
    """Test that the files can be set for the images."""
    files = [b"image data 1", b"image data 2"]

    images = [
        ImageInput(content=b"file"),
        ImageInput(numpy=b"file"),
    ]

    image_list_input = ImageInputList(images)
    image_list_input.set_files(files)

    assert image_list_input[0].content == files[0]
    assert image_list_input[1].numpy == files[1]


def test_imagelistinput_non_empty():
    """Test that ImageInputList must not be empty."""
    with pytest.raises(ValidationError):
        ImageInputList([])



from typing import TYPE_CHECKING

from aana.exceptions.core import BaseException

if TYPE_CHECKING:
    from aana.core.models.audio import Audio
    from aana.core.models.image import Image
    from aana.core.models.video import Video


__all__ = [
    "ImageReadingException",
    "AudioReadingException",
    "DownloadException",
    "VideoException",
    "VideoReadingException",
    "VideoTooLongException",
]


class ImageReadingException(BaseException):
    """Exception raised when there is an error reading an image.

    Attributes:
        image (Image): the image that caused the exception
    """

    def __init__(self, image: "Image"):
        """Initialize the exception.

        Args:
            image (Image): the image that caused the exception
        """
        super().__init__(image=image)
        self.image = image

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.image,))


class AudioReadingException(BaseException):
    """Exception raised when there is an error reading an audio.

    Attributes:
        audio (Audio): the audio that caused the exception
    """

    def __init__(self, audio: "Audio"):
        """Initialize the exception.

        Args:
            audio (Audio): the audio that caused the exception
        """
        super().__init__(audio=audio)
        self.audio = audio

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.audio,))


class DownloadException(BaseException):
    """Exception raised when there is an error downloading a file.

    Attributes:
        url (str): the URL of the file that caused the exception
    """

    def __init__(self, url: str, msg: str = ""):
        """Initialize the exception.

        Args:
            url (str): the URL of the file that caused the exception
            msg (str): the error message
        """
        super().__init__(url=url)
        self.url = url
        self.msg = msg

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.url, self.msg))


class VideoException(BaseException):
    """Exception raised when working with videos.

    Attributes:
        video (Video): the video that caused the exception
    """

    def __init__(self, video: "Video"):
        """Initialize the exception.

        Args:
            video (Video): the video that caused the exception
        """
        super().__init__(video=video)
        self.video = video

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.video,))


class VideoReadingException(VideoException):
    """Exception raised when there is an error reading a video.

    Attributes:
        video (Video): the video that caused the exception
    """

    pass


class VideoTooLongException(BaseException):
    """Exception raised when the video is too long.

    Attributes:
        video (Video): the video that caused the exception
        video_len (float): the length of the video in seconds
        max_len (float): the maximum allowed length of the video in seconds
    """

    def __init__(self, video: "Video", video_len: float, max_len: float):
        """Initialize the exception.

        Args:
            video (Video): the video that caused the exception
            video_len (float): the length of the video in seconds
            max_len (float): the maximum allowed length of the video in seconds
        """
        super().__init__(video=video, video_len=video_len, max_len=max_len)
        self.video = video
        self.video_len = video_len
        self.max_len = max_len

    def __reduce__(self):
        """Used for pickling."""
        return (self.__class__, (self.video, self.video_len, self.max_len))



from collections import defaultdict

from aana.core.models.asr import (
    AsrSegment,
    AsrWord,
)
from aana.core.models.speaker import SpeakerDiarizationSegment
from aana.core.models.time import TimeInterval

# Define sentence ending punctuations to split segments at sentence endings:
sentence_ending_punctuations = ".?!"


class PostProcessingForDiarizedAsr:
    """Class to handle post-processing for diarized ASR output by combining diarization and transcription segments.

    The post-processing involves assigning speaker labels to transcription segments and words, aligning speakers
    with punctuation, optionally merging homogeneous speaker segments, and reassigning confidence information to the segments.

    Attributes:
        diarized_segments (list[SpeakerDiarizationSegment]): Contains speaker diarization segments.
        transcription_segments (list[AsrSegment]): Transcription segments.
        merge (bool): Whether to merge the same speaker segments in the final output.
    """

    @classmethod
    def process(
        cls,
        diarized_segments: list[SpeakerDiarizationSegment],
        transcription_segments: list[AsrSegment],
        merge: bool = False,
    ) -> list[AsrSegment]:
        """Executes the post-processing pipeline that combines diarization and transcription segments.

        This method performs the following steps:
        1. Assign speaker labels to each segment and word in the transcription based on the diarization output.
        2. Align speakers with punctuation.
        3. Create new transcription segments by combining the speaker-labeled words.
        4. Optionally, merge consecutive speaker segments.
        5. Add confidence and no_speech_confidence to the new segments.

        Args:
            diarized_segments (list[SpeakerDiarizationSegment]): Contains speaker diarization segments.
            transcription_segments (list[AsrSegment]): Transcription segments.
            merge (bool): If True, merges consecutive speaker segments in the final output. Defaults to False.

        Returns:
            list[AsrSegment]: Updated transcription segments with speaker information per segment and word.
        """
        # intro: validation checks
        if not transcription_segments or not diarized_segments:
            return transcription_segments

        # Check if inputs are valid:
        for segment in transcription_segments:
            if segment.text and not segment.words:
                raise ValueError("Word-level timestamps are required for diarized ASR.")  # noqa: TRY003

        # 1. Assign speaker labels to each segment and word
        speaker_labelled_transcription = cls._assign_word_speakers(
            diarized_segments, transcription_segments
        )

        # 2. Align speakers with punctuation
        word_speaker_mapping = cls._align_with_punctuation(
            speaker_labelled_transcription
        )

        # 3. Create new transcription segments with speaker information
        segments = cls._create_speaker_segments(word_speaker_mapping)

        # 4. Assign confidence variables to the new segments
        segments = cls._add_segment_variables(segments, transcription_segments)

        # Optional: Merge consecutive speaker segments
        if merge:
            segments = cls._combine_homeogeneous_speaker_asr_segments(segments)

        return segments

    @classmethod
    def _create_speaker_segments(
        cls, word_list: list[AsrWord], max_words_per_segment: int = 50
    ) -> list[AsrSegment]:
        """Creates speaker segments from a list of words with speaker annotations.

        Args:
            word_list (list[AsrWord]): A list of words with associated speaker and timing details.
            max_words_per_segment (int): The maximum number of words per segment. If the segment exceeds this,
                                        it will be split at previous or next sentence-ending punctuation.

        Returns:
            list[AsrSegment]: A list of segments where each segment groups words spoken by the same speaker.
        """
        if not word_list:
            return []

        current_speaker = None
        current_segment = None
        final_segments: list[AsrSegment] = []
        word_count = 0

        for word_info in word_list:
            speaker = word_info.speaker or current_speaker

            # Handle speaker change
            if speaker != current_speaker:
                if current_segment:
                    final_segments.append(current_segment)
                current_segment = cls._create_new_segment(word_info, speaker)
                current_speaker = speaker
                word_count = 1
            else:
                if current_segment:
                    # Handle word count and punctuation splitting
                    (
                        current_segment,
                        word_count,
                    ) = cls._split_segment_on_length_punctuation(
                        current_segment,
                        word_info,
                        word_count,
                        max_words_per_segment,
                        final_segments,
                    )

        # Add the final segment if it exists
        if current_segment and current_segment.words:
            final_segments.append(current_segment)

        return final_segments

    @classmethod
    def _align_with_punctuation(
        cls, transcription_segments: list[AsrSegment], max_words_in_sentence: int = 50
    ) -> list[AsrWord]:
        """Aligns speaker labels with sentence boundaries defined by punctuation.

        Args:
            transcription_segments (list[AsrSegment]): transcription segments with speaker information.
            max_words_in_sentence (int): Maximum number of words allowed in a sentence.

        Returns:
            word_speaker_mapping: (list[AsrWord]): Realigned word-speaker mappings.
        """
        new_segments = [segment.words for segment in transcription_segments]
        word_speaker_mapping = [word for segment in new_segments for word in segment]
        words_list = [item.word for item in word_speaker_mapping]
        speaker_list = [item.speaker for item in word_speaker_mapping]

        # Fill missing speaker labels by finding the nearest speaker
        for i, item in enumerate(word_speaker_mapping):
            if item.speaker is None:
                item.speaker = cls._find_nearest_speaker(
                    i, word_speaker_mapping, reverse=i > 0
                )
                speaker_list[i] = item.speaker

        # Align speakers with sentence boundaries
        k = 0
        while k < len(word_speaker_mapping):
            if (
                k < len(word_speaker_mapping) - 1
                and speaker_list[k] != speaker_list[k + 1]
                and words_list[k][-1] not in sentence_ending_punctuations
            ):
                left_idx = cls._get_first_word_idx_of_sentence(
                    k, words_list, speaker_list, max_words_in_sentence
                )
                right_idx = cls._get_last_word_idx_of_sentence(
                    k, words_list, max_words_in_sentence - (k - left_idx)
                )

                if min(left_idx, right_idx) == -1:
                    k += 1
                    continue

                spk_labels = speaker_list[left_idx : right_idx + 1]
                mod_speaker = max(set(spk_labels), key=spk_labels.count)

                if spk_labels.count(mod_speaker) >= len(spk_labels) // 2:
                    speaker_list[left_idx : right_idx + 1] = [mod_speaker] * (
                        right_idx - left_idx + 1
                    )
                    k = right_idx

            k += 1

        # Realign the speaker labels in the original word_speaker_mapping
        for i, item in enumerate(word_speaker_mapping):
            item.speaker = speaker_list[i]

        return word_speaker_mapping

    @classmethod
    def _add_segment_variables(
        cls, segments: list[AsrSegment], transcription_segments: list[AsrSegment]
    ) -> list[AsrSegment]:
        """Adds confidence and no_speech_confidence variables to each segment.

        Args:
            segments (list[AsrSegment]): A list of segments to which the confidence values will be added.
            transcription_segments (list[AsrSegment]): The original transcription segments to help determine segment confidence.

        Returns:
            list[AsrSegment]: Segments with confidence and no_speech_confidence added.
        """
        for segment in segments:
            confidence, no_speech_confidence = cls._determine_major_segment_confidence(
                segment, transcription_segments
            )
            segment.confidence = confidence
            segment.no_speech_confidence = no_speech_confidence
        return segments

    @classmethod
    def _split_segment_on_length_punctuation(
        cls,
        current_segment: AsrSegment,
        word_info: AsrWord,
        word_count: int,
        max_words_per_segment: int,
        final_segments: list[AsrSegment],
    ) -> tuple[AsrSegment, int]:
        """Splits segments based on length and sentence-ending punctuation.

        Args:
            current_segment (AsrSegment): The current speaker segment being processed.
            word_info (AsrWord): Word information containing timing and text.
            word_count (int): The current word count in the segment.
            max_words_per_segment (int): Maximum number of words allowed in a segment before splitting.
            final_segments (list[AsrSegment]): List of segments to which the completed segment will be added.

        Returns:
            Tuple[AsrSegment, int]: The updated segment and word count.
        """
        # Check if word count exceeds the limit and if punctuation exists to split
        if word_count >= max_words_per_segment and any(
            p in word_info.word for p in sentence_ending_punctuations
        ):
            # update current segment and then append it
            current_segment.time_interval.end = word_info.time_interval.end
            current_segment.text += f"{word_info.word}"
            current_segment.words.append(word_info)
            final_segments.append(current_segment)
            current_segment = cls._create_new_segment(
                word_info, current_segment.speaker, is_empty=True
            )
            word_count = 0  # Reset word count

        else:
            # Append word to the current segment
            current_segment.time_interval.end = word_info.time_interval.end
            current_segment.text += f"{word_info.word}"
            current_segment.words.append(word_info)
            word_count += 1

        return current_segment, word_count

    @staticmethod
    def _assign_word_speakers(
        diarized_segments: list[SpeakerDiarizationSegment],
        transcription_segments: list[AsrSegment],
        fill_nearest: bool = False,
    ) -> list[AsrSegment]:
        """Assigns speaker labels to each segment and word in the transcription based on diarized output.

        Args:
            diarized_segments (list[SpeakerDiarizationSegment]): Contains speaker diarization segments.
            transcription_segments (list[AsrSegment]): Transcription data with segments, text, and language_info.
            fill_nearest (bool): If True, assigns the closest speaker even if there's no positive overlap. Default is False.

        Returns:
            transcription_segments (list[AsrSegment]): Transcription updated in-place with the assigned speaker labels.
        """

        def get_speaker_for_interval(
            sd_segments: list[SpeakerDiarizationSegment],
            start_time: float,
            end_time: float,
            fill_nearest: bool,
        ) -> str | None:
            """Determines the speaker for a given time interval based on diarized segments.

            Args:
                sd_segments (list[SpeakerDiarizationSegment]): List of speaker diarization segments.
                start_time (float): Start time of the interval.
                end_time (float): End time of the interval.
                fill_nearest (bool): If True, selects the closest speaker even with no overlap.

            Returns:
                str | None: The identified speaker label, or None if no speaker is found.
            """
            overlaps = []

            for sd_segment in sd_segments:
                interval_start = sd_segment.time_interval.start
                interval_end = sd_segment.time_interval.end

                # Calculate overlap duration
                overlap_start = max(start_time, interval_start)
                overlap_end = min(end_time, interval_end)
                overlap_duration = max(0.0, overlap_end - overlap_start)

                if overlap_duration > 0 or fill_nearest:
                    distance = float(
                        min(
                            abs(start_time - interval_end),
                            abs(end_time - interval_start),
                        )
                    )

                    overlaps.append(
                        {
                            "speaker": sd_segment.speaker,
                            "overlap_duration": overlap_duration,
                            "distance": distance,
                        }
                    )

            if not overlaps:
                return None
            else:
                # Select the speaker with the maximum overlap duration (or minimal distance)
                best_match = max(
                    overlaps,
                    key=lambda x: (x["overlap_duration"], -x["distance"])
                    if not fill_nearest
                    else (-x["distance"]),
                )
                return best_match["speaker"]

        for segment in transcription_segments:
            # Assign speaker to segment
            segment.speaker = get_speaker_for_interval(
                diarized_segments,
                segment.time_interval.start,
                segment.time_interval.end,
                fill_nearest,
            )

            # Assign speakers to words within the segment
            if segment.words:
                for word in segment.words:
                    word.speaker = get_speaker_for_interval(
                        diarized_segments,
                        word.time_interval.start,
                        word.time_interval.end,
                        fill_nearest,
                    )

        return transcription_segments

    @staticmethod
    def _get_first_word_idx_of_sentence(
        word_idx: int,
        word_list: list[str],
        speaker_list: list[str | None],
        max_words: int,
    ) -> int:
        """Get the index of the first word of the sentence in the given range."""
        left_idx = word_idx
        while (
            left_idx > 0
            and word_idx - left_idx < max_words
            and speaker_list[left_idx - 1] == speaker_list[left_idx]
            and word_list[left_idx - 1][-1] not in sentence_ending_punctuations
        ):
            left_idx -= 1

        return (
            left_idx
            if left_idx == 0
            or word_list[left_idx - 1][-1] in sentence_ending_punctuations
            else -1
        )

    @staticmethod
    def _get_last_word_idx_of_sentence(
        word_idx: int, word_list: list[str], max_words: int
    ) -> int:
        """Get the index of the last word of the sentence in the given range."""
        right_idx = word_idx
        while (
            right_idx < len(word_list)
            and right_idx - word_idx < max_words
            and word_list[right_idx][-1] not in sentence_ending_punctuations
        ):
            right_idx += 1

        return (
            right_idx
            if right_idx == len(word_list) - 1
            or word_list[right_idx][-1] in sentence_ending_punctuations
            else -1
        )

    @staticmethod
    def _find_nearest_speaker(
        index: int, word_speaker_mapping: list[AsrWord], reverse: bool = False
    ) -> str | None:
        """Find the nearest speaker label in the word_speaker_mapping either forward or backward.

        Args:
            index (int): The index to start searching from.
            word_speaker_mapping (list[AsrWord]): List of word-speaker mappings.
            reverse (bool): Search backwards if True; forwards if False. Default is False.

        Returns:
            str | None: The nearest speaker found or None if not found.
        """
        step = -1 if reverse else 1
        for i in range(index, len(word_speaker_mapping) if not reverse else -1, step):
            if word_speaker_mapping[i].speaker:
                return word_speaker_mapping[i].speaker
        return None

    @staticmethod
    def _create_new_segment(
        word_info: AsrWord, speaker: str | None, is_empty: bool = False
    ) -> AsrSegment:
        """Creates a new segment based on word information.

        Args:
            word_info (AsrWord): The word information containing text, timing, etc.
            speaker (str | None): The speaker associated with this word.
            is_empty (bool): If True, creates an empty segment (for punctuation-only segments).

        Returns:
            AsrSegment: A new segment with the provided word information and speaker details.
        """
        return AsrSegment(
            time_interval=TimeInterval(
                start=word_info.time_interval.start
                if not is_empty
                else word_info.time_interval.end,
                end=word_info.time_interval.end,
            ),
            text=word_info.word if not is_empty else "",
            speaker=speaker,
            words=[word_info] if not is_empty else [],
            confidence=None,
            no_speech_confidence=None,
        )

    @staticmethod
    def _determine_major_segment_confidence(
        segment: AsrSegment, transcription_segments: list[AsrSegment]
    ) -> tuple[float | None, float | None]:
        """Determines the confidence and no_speech_confidence based on the major segment (which contributes the most time or words).

        Args:
            segment (AsrSegment): New ASR segment.
            transcription_segments (list[AsrSegment]): Original transcription segments with confidence.

        Returns:
            tuple[Optional[float], Optional[float]]: Confidence and no_speech_confidence from the major segment.
        """

        def find_closest_segment(
            word_start: float, word_end: float
        ) -> AsrSegment | None:
            """Finds the closest segment in the transcription for the given word start and end times."""
            closest_segment = min(
                transcription_segments,
                key=lambda segment: abs(segment.time_interval.start - word_start)
                + abs(segment.time_interval.end - word_end),
                default=None,
            )
            return closest_segment

        def update_segment_contribution(
            contributions: dict, segment: AsrSegment, word_duration: float
        ) -> None:
            """Updates the contribution data for the given segment."""
            segment_id = id(segment)
            if segment_id not in contributions:
                contributions[segment_id] = {
                    "segment": segment,
                    "contribution_time": 0.0,
                    "word_count": 0,
                }
            contributions[segment_id]["contribution_time"] += word_duration
            contributions[segment_id]["word_count"] += 1

        segment_contributions: defaultdict = defaultdict(
            lambda: {"segment": None, "contribution_time": 0.0, "word_count": 0}
        )

        for word in segment.words:
            word_start, word_end = word.time_interval.start, word.time_interval.end
            word_duration = word_end - word_start

            closest_segment = find_closest_segment(word_start, word_end)

            if closest_segment:
                update_segment_contribution(
                    segment_contributions, closest_segment, word_duration
                )

        if not segment_contributions:
            return None, None

        # Determine the segment with the highest word count or contribution time
        major_segment_data = max(
            segment_contributions.values(),
            key=lambda data: data[
                "word_count"
            ],  # Change this to 'contribution_time' if needed
        )

        major_segment = major_segment_data["segment"]
        return major_segment.confidence, major_segment.no_speech_confidence

    @staticmethod
    def _combine_homeogeneous_speaker_asr_segments(
        segments: list[AsrSegment],
    ) -> list[AsrSegment]:
        """Merges consecutive segments that have the same speaker into a single segment.

        Args:
            segments (list[AsrSegment]): The initial list of segments.

        Returns:
            merged_segments (list[AsrSegment]): A new list of merged segments.
        """
        if not segments:
            return []

        merged_segments: list[AsrSegment] = []
        mapping: defaultdict[str, str] = defaultdict(str)

        current_segment = segments[0]
        speaker_counter = 0

        for next_segment in segments[1:]:
            if next_segment.speaker == current_segment.speaker:
                # Merge segments
                current_segment.time_interval.end = next_segment.time_interval.end
                current_segment.text += f" {next_segment.text}"
                current_segment.words.extend(next_segment.words)
            else:
                # Assign unique speaker labels and finalize the current segment
                if current_segment.speaker:
                    current_segment.speaker = mapping.setdefault(
                        current_segment.speaker, f"SPEAKER_{speaker_counter:02d}"
                    )
                    for word in current_segment.words:
                        word.speaker = current_segment.speaker
                    merged_segments.append(current_segment)
                    current_segment = next_segment
                    speaker_counter += 1

        # Handle the last segment
        if current_segment.speaker:
            current_segment.speaker = mapping.setdefault(
                current_segment.speaker, f"SPEAKER_{speaker_counter:02d}"
            )
            for word in current_segment.words:
                word.speaker = current_segment.speaker
            merged_segments.append(current_segment)

        return merged_segments


def combine_homogeneous_speaker_diarization_segments(
    diarized_segments: list[SpeakerDiarizationSegment],
) -> list[SpeakerDiarizationSegment]:
    """Combines same speaker segments in a diarization output.

    Speaker diarization model occationally produce overlapping chunks or segments belonging to the same speaker.
    This method combines segments with the same speaker into homogeneous speaker segments, ensuring no overlapping times.

    Args:
        diarized_segments (list(SpeakerDiarizationSegment)): Input with segments that may have overlapping times.

    Returns:
        list(SpeakerDiarizationSegment): Output with combined homogeneous speaker segments.
    """
    combined_segments: list = []
    current_speaker = None
    current_segment = None

    for segment in sorted(diarized_segments, key=lambda x: x.time_interval.start):
        speaker = segment.speaker

        # If there's a speaker change or current_segment is None, finalize current and start a new one
        if current_speaker != speaker:
            # Finalize the current segment if it exists
            if current_segment:
                combined_segments.append(current_segment)

            current_speaker = speaker

            # Start a new segment for the current speaker
            current_segment = SpeakerDiarizationSegment(
                time_interval=TimeInterval(
                    start=segment.time_interval.start, end=segment.time_interval.end
                ),
                speaker=current_speaker,
            )
        else:
            if current_segment:
                # Extend the current segment for the same speaker
                # Ensure there is no overlap; take the maximum of current end and incoming end
                current_segment.time_interval.end = max(
                    current_segment.time_interval.end, segment.time_interval.end
                )

        # Adjust the start of the next segment if there's any overlap
        if (
            current_segment
            and len(combined_segments) > 0
            and combined_segments[-1].time_interval.end
            > current_segment.time_interval.start
        ):
            current_segment.time_interval.start = combined_segments[
                -1
            ].time_interval.end

    # Add the last segment if it exists
    if current_segment:
        combined_segments.append(current_segment)

    return combined_segments



from pathlib import Path

import cv2
import numpy as np

from aana.core.libraries.image import AbstractImageLibrary

__all__ = ["OpenCVWrapper"]


class OpenCVWrapper(AbstractImageLibrary):
    """Wrapper class for OpenCV functions."""

    @classmethod
    def read_file(cls, path: Path) -> np.ndarray:
        """Read a file using OpenCV.

        Args:
            path (Path): The path of the file to read.

        Returns:
            np.ndarray: The file as a numpy array in RGB format.
        """
        img = cv2.imread(str(path))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img

    @classmethod
    def read_from_bytes(cls, content: bytes) -> np.ndarray:
        """Read bytes using OpenCV.

        Args:
            content (bytes): The content of the file to read.

        Returns:
            np.ndarray: The file as a numpy array in RGB format.
        """
        img = cv2.imdecode(np.frombuffer(content, np.uint8), cv2.IMREAD_COLOR)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img

    @classmethod
    def write_file(cls, path: Path, img: np.ndarray):
        """Write a file using OpenCV.

        Args:
            path (Path): The path of the file to write.
            img (np.ndarray): The image to write.
        """
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        cv2.imwrite(str(path), img)

    @classmethod
    def write_to_bytes(cls, img: np.ndarray) -> bytes:
        """Write bytes using OpenCV.

        Args:
            img (np.ndarray): The image to write.

        Returns:
            bytes: The image as bytes.
        """
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        _, buffer = cv2.imencode(".bmp", img)
        return buffer.tobytes()



# ruff: noqa: S101, S113
from typing import Annotated, TypedDict

from pydantic import Field
from ray import serve

from aana.api.api_generation import Endpoint
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment
from aana.sdk import AanaSDK


@serve.deployment
class Lowercase(BaseDeployment):
    """Ray deployment that returns the lowercase version of a text."""

    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        return {"text": [t.lower() for t in text]}


TextList = Annotated[list[str], Field(description="List of text to lowercase.")]


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: list[str]


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.lowercase_handle = await AanaDeploymentHandle.create(
            "lowercase_deployment"
        )
        await super().initialize()

    async def run(self, text: TextList) -> LowercaseEndpointOutput:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        lowercase_output = await self.lowercase_handle.lower(text=text)
        return {"text": lowercase_output["text"]}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    }
]

endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseEndpoint,
    }
]

aana_app = AanaSDK(name="lowercase_app")

for deployment in deployments:
    aana_app.register_deployment(
        name=deployment["name"],
        instance=deployment["instance"],
    )

for endpoint in endpoints:
    aana_app.register_endpoint(
        name=endpoint["name"],
        path=endpoint["path"],
        summary=endpoint["summary"],
        endpoint_cls=endpoint["endpoint_cls"],
    )



# ruff: noqa: S101
from importlib import resources
from pathlib import Path

import pytest
from pydantic import ValidationError

from aana.core.models.video import VideoInput, VideoInputList


@pytest.fixture
def mock_download_file(mocker):
    """Mock download_file."""
    mock = mocker.patch("aana.core.models.media.download_file", autospec=True)
    path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
    content = path.read_bytes()
    mock.return_value = content
    return mock


def test_new_videoinput_success():
    """Test that VideoInput can be created successfully."""
    video_input = VideoInput(path="video.mp4")
    assert video_input.path == "video.mp4"

    video_input = VideoInput(url="http://example.com/video.mp4")
    assert video_input.url == "http://example.com/video.mp4"

    video_input = VideoInput(content=b"file")
    assert video_input.content == b"file"


def test_videoinput_invalid_media_id():
    """Test that VideoInput can't be created if media_id is invalid."""
    with pytest.raises(ValidationError):
        VideoInput(path="video.mp4", media_id="")


@pytest.mark.parametrize(
    "url",
    [
        "domain",
        "domain.com",
        "http://",
        "www.domain.com",
        "/subdomain",
        "../subdomain",
        "",
    ],
)
def test_videoinput_invalid_url(url):
    """Test that VideoInput can't be created if url is invalid."""
    with pytest.raises(ValidationError):
        VideoInput(url=url)


def test_videoinput_check_only_one_field():
    """Test that exactly one of 'path', 'url', or 'content' is provided."""
    fields = {
        "path": "video.mp4",
        "url": "http://example.com/video.mp4",
        "content": b"file",
    }

    # check all combinations of two fields
    for field1 in fields:
        for field2 in fields:
            if field1 != field2:
                with pytest.raises(ValidationError):
                    VideoInput(**{field1: fields[field1], field2: fields[field2]})

    # check all combinations of three fields
    with pytest.raises(ValidationError):
        VideoInput(**fields)

    # check that no fields is also invalid
    with pytest.raises(ValidationError):
        VideoInput()


def test_videoinput_set_file():
    """Test that the file can be set for the video."""
    file_content = b"video data"
    video_input = VideoInput(content=b"file")
    video_input.set_file(file_content)
    assert video_input.content == file_content

    # If 'content' is not set to 'file',
    # an error should be raised.
    video_input = VideoInput(path="video.mp4")
    with pytest.raises(ValueError):
        video_input.set_file(file_content)


def test_videoinput_set_files():
    """Test that the files can be set for the video."""
    files = [b"video data"]

    video_input = VideoInput(content=b"file")
    video_input.set_files(files)
    assert video_input.content == files[0]

    # If 'content' is not set to 'file',
    # an error should be raised.
    video_input = VideoInput(path="video.mp4")
    with pytest.raises(ValueError):
        video_input.set_files(files)

    # If the number of files is not 1,
    # an error should be raised.
    files = [b"video data", b"another video data"]
    video_input = VideoInput(content=b"file")
    with pytest.raises(ValidationError):
        video_input.set_files(files)

    files = []
    video_input = VideoInput(content=b"file")
    with pytest.raises(ValidationError):
        video_input.set_files(files)


def test_videoinput_convert_input_to_object(mock_download_file):
    """Test that VideoInput can be converted to Video."""
    path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
    video_input = VideoInput(path=str(path))
    try:
        video_object = video_input.convert_input_to_object()
        assert video_object.path == path
    finally:
        video_object.cleanup()

    url = "http://example.com/squirrel.mp4"
    video_input = VideoInput(url=url)
    try:
        video_object = video_input.convert_input_to_object()
        assert video_object.url == url
    finally:
        video_object.cleanup()

    content = Path(path).read_bytes()
    video_input = VideoInput(content=content)
    try:
        video_object = video_input.convert_input_to_object()
        assert video_object.content == content
    finally:
        video_object.cleanup()


def test_videoinputlist():
    """Test that VideoInputList can be created successfully."""
    videos = [
        VideoInput(path="video.mp4"),
        VideoInput(url="http://example.com/video.mp4"),
        VideoInput(content=b"file"),
    ]

    video_list_input = VideoInputList(root=videos)
    assert video_list_input.root == videos
    assert len(video_list_input) == len(videos)
    assert video_list_input[0] == videos[0]
    assert video_list_input[1] == videos[1]
    assert video_list_input[2] == videos[2]


def test_videoinputlist_set_files():
    """Test that the files can be set for the video list."""
    files = [b"video data", b"another video data"]

    videos = [
        VideoInput(content=b"file"),
        VideoInput(content=b"file"),
    ]

    video_list_input = VideoInputList(root=videos)
    video_list_input.set_files(files)
    assert video_list_input[0].content == files[0]
    assert video_list_input[1].content == files[1]

    # If the number of files is not the same as the number of videos,
    # an error should be raised.
    files = [b"video data", b"another video data", b"yet another video data"]
    video_list_input = VideoInputList(root=videos)
    with pytest.raises(ValidationError):
        video_list_input.set_files(files)

    files = []
    video_list_input = VideoInputList(root=videos)
    with pytest.raises(ValidationError):
        video_list_input.set_files(files)


def test_videoinputlist_non_empty():
    """Test that videoinputlist must not be empty."""
    with pytest.raises(ValidationError):
        VideoInputList(root=[])



# ruff: noqa: S101
import json
from collections import defaultdict
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import rapidfuzz
import requests
from pydantic import ValidationError

from aana.api.api_generation import Endpoint
from aana.configs.settings import settings
from aana.core.models.base import pydantic_to_dict
from aana.sdk import AanaSDK
from aana.tests.const import ALLOWED_LEVENSTEIN_ERROR_RATE
from aana.utils.json import jsonify


def is_gpu_available() -> bool:
    """Check if a GPU is available.

    Returns:
        bool: True if a GPU is available, False otherwise.
    """
    import torch

    # TODO: find the way to check if GPU is available without importing torch
    return torch.cuda.is_available()


def round_floats(obj, decimals=2):
    """Round floats in a nested structure."""
    if isinstance(obj, float):
        return round(obj, decimals)
    elif isinstance(obj, defaultdict):
        return defaultdict(
            obj.default_factory, {k: round_floats(v, decimals) for k, v in obj.items()}
        )
    elif isinstance(obj, Mapping):
        return type(obj)((k, round_floats(v, decimals)) for k, v in obj.items())
    elif isinstance(obj, Sequence) and not isinstance(obj, str):  # noqa: SIM114
        return type(obj)(round_floats(x, decimals) for x in obj)
    elif isinstance(obj, set):
        return type(obj)(round_floats(x, decimals) for x in obj)
    elif hasattr(obj, "__dict__"):
        obj.__dict__.update(
            (k, round_floats(v, decimals)) for k, v in obj.__dict__.items()
        )
        return obj
    else:
        return obj


def compare_texts(
    expected_text: str,
    text: str,
    allowed_error_rate: float = ALLOWED_LEVENSTEIN_ERROR_RATE,
):
    """Compare two texts using Levenshtein distance.

    The error rate is allowed to be less than ALLOWED_LEVENSTEIN_ERROR_RATE.

    Args:
        expected_text (str): the expected text
        text (str): the actual texts
        allowed_error_rate (float): the allowed error rate (default: ALLOWED_LEVENSTEIN_ERROR_RATE)

    Raises:
        AssertionError: if the error rate is too high
    """
    dist = rapidfuzz.distance.Levenshtein.distance(text, expected_text)
    assert dist < len(expected_text) * allowed_error_rate, (
        expected_text,
        text,
        dist,
    )


def compare_results(
    expected_result: dict[str, Any],
    result: dict[str, Any],
    allowed_error_rate: float = ALLOWED_LEVENSTEIN_ERROR_RATE,
    apply_float_rounding=True,
    round_decimals=2,
):
    """Compare expected and actual results by jsonifying them and comparing the strings using Levenshtein distance.

    Args:
        expected_result (dict): the expected result
        result (dict): the actual result
        allowed_error_rate (float): the allowed error rate (default: ALLOWED_LEVENSTEIN_ERROR_RATE)
        apply_float_rounding (bool): whether to round floats (default: True)
        round_decimals (int): the number of decimals to round floats to (default: 2)

    Raises:
        AssertionError: if the error rate is too high
    """
    if apply_float_rounding:
        expected_json = jsonify(round_floats(expected_result, round_decimals))
        result_json = jsonify(round_floats(result, round_decimals))
    else:
        expected_json = jsonify(expected_result)
        result_json = jsonify(result)
    compare_texts(expected_json, result_json, allowed_error_rate)


def verify_deployment_results(
    expected_output_path: Path,
    results: dict[str, Any],
    allowed_error_rate: float = ALLOWED_LEVENSTEIN_ERROR_RATE,
):
    """Verify the output of a deployment call against an expected output file.

    If the expected output file doesn't exist and the SAVE_EXPECTED_OUTPUT
    environment variable is set, save the current results as the expected output.
    Creates parent directories if they don't exist.

    Args:
        expected_output_path (Path): Path to the expected output file.
        results (dict): The actual results.
        allowed_error_rate (float): The allowed error rate (default: ALLOWED_LEVENSTEIN_ERROR_RATE)
    """
    # Convert pydantic models to dictionaries before comparison
    results = pydantic_to_dict(results)

    if not expected_output_path.exists():
        if settings.test.save_expected_output:
            # Ensure the parent directory exists
            expected_output_path.parent.mkdir(parents=True, exist_ok=True)

            with expected_output_path.open("w") as f:
                json.dump(results, f, indent=2)
            print(f"Saved results as expected output to: {expected_output_path}")
            return
        else:
            raise FileNotFoundError(  # noqa: TRY003
                f"Expected output not found: {expected_output_path}"
            )

    with expected_output_path.open() as f:
        expected_output = json.load(f)

    compare_results(expected_output, results, allowed_error_rate)


def send_api_request(
    endpoint: Endpoint,
    app: AanaSDK,
    data: dict[str, Any],
    timeout: int = 30,
) -> dict[str, Any] | list[dict[str, Any]]:
    """Call an endpoint, handling both streaming and non-streaming responses."""
    url = f"http://localhost:{app.port}{endpoint.path}"
    payload = {"body": json.dumps(data)}

    if endpoint.is_streaming_response():
        output = []
        with requests.post(url, data=payload, timeout=timeout, stream=True) as r:
            for chunk in r.iter_content(chunk_size=None):
                chunk_output = json.loads(chunk.decode("utf-8"))
                output.append(chunk_output)
                if "error" in chunk_output:
                    return [chunk_output]
        return output
    else:
        response = requests.post(url, data=payload, timeout=timeout)
        return response.json()


def verify_output(
    endpoint: Endpoint,
    response: dict[str, Any] | list[dict[str, Any]],
    expected_error: str | None = None,
) -> None:
    """Verify the output of an endpoint call."""
    is_streaming = endpoint.is_streaming_response()
    ResponseModel = endpoint.get_response_model()
    if expected_error:
        error = response[0]["error"] if is_streaming else response["error"]
        assert error == expected_error, response
    else:
        try:
            if is_streaming:
                for item in response:
                    ResponseModel.model_validate(item, strict=True)
            else:
                ResponseModel.model_validate(response, strict=True)
        except ValidationError as e:
            raise AssertionError(  # noqa: TRY003
                f"Validation failed. Errors:\n{e}\n\nResponse: {response}"
            ) from e



import inspect
from functools import wraps
from typing import Any

from aana.exceptions.runtime import InferenceException


def exception_handler(func):
    """AanaDeploymentHandle decorator to catch exceptions and store them in the deployment for health check purposes.

    Args:
        func (function): The function to decorate.

    Returns:
        function: The decorated function
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        self.num_requests_since_last_health_check += 1
        try:
            return await func(self, *args, **kwargs)
        except Exception as e:
            self.raised_exceptions.append(e)
            raise

    @wraps(func)
    async def wrapper_generator(self, *args, **kwargs):
        self.num_requests_since_last_health_check += 1
        try:
            async for item in func(self, *args, **kwargs):
                yield item
        except Exception as e:
            self.raised_exceptions.append(e)
            raise

    if inspect.isasyncgenfunction(func):
        return wrapper_generator
    else:
        return wrapper


class BaseDeployment:
    """Base class for all deployments.

    To create a new deployment, inherit from this class and implement the `apply_config` method
    and your custom methods like `generate`, `predict`, etc.
    """

    def __init__(self):
        """Inits to unconfigured state."""
        self.config = None
        self._configured = False
        self.num_requests_since_last_health_check = 0
        self.raised_exceptions = []
        self.restart_exceptions = [InferenceException]

    async def reconfigure(self, config: dict[str, Any]):
        """Reconfigure the deployment.

        The method is called when the deployment is updated.
        """
        self.config = config
        await self.apply_config(config)
        self._configured = True

    async def check_health(self):
        """Check the health of the deployment.

        Raises:
            Raises the exception that caused the deployment to be unhealthy.
        """
        raised_restart_exceptions = [
            exception
            for exception in self.raised_exceptions
            if exception.__class__ in self.restart_exceptions
        ]
        # Restart the deployment if more than 50% of the requests raised restart exceptions
        if self.num_requests_since_last_health_check != 0:
            ratio_restart_exceptions = (
                len(raised_restart_exceptions)
                / self.num_requests_since_last_health_check
            )
            if ratio_restart_exceptions > 0.5:
                raise raised_restart_exceptions[0]

        self.raised_exceptions = []
        self.num_requests_since_last_health_check = 0

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        This method is called when the deployment is created or updated.

        Define the logic to load the model and configure it here.

        Args:
            config (dict): the configuration
        """
        raise NotImplementedError

    async def get_methods(self) -> dict:
        """Returns the methods of the deployment.

        Returns:
            dict: the methods of the deployment with annotations and docstrings
        """
        cls = self.__class__
        methods = inspect.getmembers(cls, predicate=inspect.isfunction)
        methods_info = {}
        for name, method in methods:
            # Skip private methods
            if name.startswith("_"):
                continue
            # Skip non-asynchronous methods
            if not (
                inspect.iscoroutinefunction(method)
                or inspect.isasyncgenfunction(method)
            ):
                continue

            methods_info[name] = {}
            if method.__annotations__:
                methods_info[name]["annotations"] = method.__annotations__
            if method.__doc__:
                methods_info[name]["doc"] = method.__doc__
        return methods_info



# ruff: noqa: S101
from importlib import resources
from pathlib import Path

import pytest

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_to_dict
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.pyannote_speaker_diarization_deployment import (
    PyannoteSpeakerDiarizationConfig,
    PyannoteSpeakerDiarizationDeployment,
)
from aana.tests.utils import verify_deployment_results

deployments = [
    (
        "sd_deployment",
        PyannoteSpeakerDiarizationDeployment.options(
            num_replicas=1,
            max_ongoing_requests=1000,
            ray_actor_options={"num_gpus": 0.05},
            user_config=PyannoteSpeakerDiarizationConfig(
                model_id=("pyannote/speaker-diarization-3.1"),
                sample_rate=16000,
            ).model_dump(mode="json"),
        ),
    )
]


@pytest.mark.parametrize("setup_deployment", deployments, indirect=True)
class TestPyannoteSpeakerDiarizationDeployment:
    """Test pyannote Speaker Diarization deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("audio_file", ["sd_sample.wav"])
    async def test_speaker_diarization(self, setup_deployment, audio_file):
        """Test pyannote Speaker Diarization."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        audio_file_name = Path(audio_file).stem
        expected_output_path = (
            resources.files("aana.tests.files.expected")
            / "sd"
            / f"{audio_file_name}.json"
        )

        path = resources.files("aana.tests.files.audios") / audio_file
        assert path.exists(), f"Audio not found: {path}"

        audio = Audio(path=path)

        output = await handle.diarize(audio=audio)
        output = pydantic_to_dict(output)

        verify_deployment_results(expected_output_path, output)



# ruff: noqa: S101, S113
import asyncio

import pytest
from ray import serve

from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException


@serve.deployment(health_check_period_s=1, health_check_timeout_s=30)
class Lowercase(BaseDeployment):
    """Ray deployment that returns the lowercase version of a text."""

    def __init__(self):
        """Initialize the deployment."""
        super().__init__()
        self.active = True

    @exception_handler
    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        if text == "inference_exception" or not self.active:
            self.active = False
            raise InferenceException(model_name="lowercase_deployment")

        return {"text": text.lower()}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    }
]


@pytest.mark.asyncio
async def test_deployment_restart(create_app):
    """Test the Ray Serve app."""
    create_app(deployments, [])

    handle = await AanaDeploymentHandle.create("lowercase_deployment")

    text = "Hello, World!"

    # test the lowercase deployment works
    response = await handle.lower(text=text)
    assert response == {"text": text.lower()}

    # Cause an InferenceException in the deployment and make it inactive.
    # After the deployment is inactive, the deployment should always raise an InferenceException.
    with pytest.raises(InferenceException):
        await handle.lower(text="inference_exception")

    # The deployment should restart and work again, wait for around 60 seconds for the deployment to restart.
    for _ in range(60):
        await asyncio.sleep(1)
        try:
            response = await handle.lower(text=text)
            if response == {"text": text.lower()}:
                break
        except:  # noqa: S110
            pass

    assert response == {"text": text.lower()}



from pathlib import Path

import numpy as np


class AbstractImageLibrary:
    """Abstract class for image libraries."""

    @classmethod
    def read_file(cls, path: Path) -> np.ndarray:
        """Read a file using the image library.

        Args:
            path (Path): The path of the file to read.

        Returns:
            np.ndarray: The file as a numpy array.
        """
        raise NotImplementedError

    @classmethod
    def read_from_bytes(cls, content: bytes) -> np.ndarray:
        """Read bytes using the image library.

        Args:
            content (bytes): The content of the file to read.

        Returns:
            np.ndarray: The file as a numpy array.
        """
        raise NotImplementedError

    @classmethod
    def write_file(cls, path: Path, img: np.ndarray):
        """Write a file using the image library.

        Args:
            path (Path): The path of the file to write.
            img (np.ndarray): The image to write.
        """
        raise NotImplementedError

    @classmethod
    def write_to_bytes(cls, img: np.ndarray) -> bytes:
        """Write bytes using the image library.

        Args:
            img (np.ndarray): The image to write.

        Returns:
            bytes: The image as bytes.
        """
        raise NotImplementedError



# ruff: noqa: S101, S113
import asyncio
import json
import os
from typing import TypedDict

import pytest
import requests
from ray import serve

from aana.api.api_generation import Endpoint
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment


@serve.deployment
class Lowercase(BaseDeployment):
    """Ray deployment that returns the lowercase version of a text."""

    async def lower(self, text: str) -> dict:
        """Lowercase the text.

        Args:
            text (str): The text to lowercase

        Returns:
            dict: The lowercase text
        """
        # kill the deployment if the text is "kill"
        if text == "kill":
            os._exit(0)
        return {"text": text.lower()}


class LowercaseEndpointOutput(TypedDict):
    """The output of the lowercase endpoint."""

    text: str


class LowercaseEndpoint(Endpoint):
    """Lowercase endpoint."""

    async def initialize(self):
        """Initialize the endpoint."""
        self.lowercase_handle = await AanaDeploymentHandle.create(
            "lowercase_deployment"
        )
        await super().initialize()

    async def run(self, text: str) -> LowercaseEndpointOutput:
        """Lowercase the text.

        Args:
            text (TextList): The list of text to lowercase

        Returns:
            LowercaseEndpointOutput: The lowercase texts
        """
        lowercase_output = await self.lowercase_handle.lower(text=text)
        return {"text": lowercase_output["text"]}


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": Lowercase,
    }
]

endpoints = [
    {
        "name": "lowercase",
        "path": "/lowercase",
        "summary": "Lowercase text",
        "endpoint_cls": LowercaseEndpoint,
    }
]


def get_status(app):
    """Get the status of the app."""
    response = requests.get(f"http://localhost:{app.port}/api/status")
    return response.json()


@pytest.mark.asyncio
async def test_status_endpoint(create_app):
    """Test SDK status endpoint."""
    aana_app = create_app(deployments, endpoints)

    status = get_status(aana_app)
    assert status["status"] == "RUNNING"

    # kill the deployment
    response = requests.post(  # noqa: ASYNC100
        f"http://localhost:{aana_app.port}/lowercase",
        data={"body": json.dumps({"text": "kill"})},
    ).json()
    assert "error" in response

    # Wait for the UNHEALTHY status for 30 seconds
    for _ in range(30):
        status = get_status(aana_app)
        if status["status"] == "UNHEALTHY":
            break
        await asyncio.sleep(1)
    assert status["status"] == "UNHEALTHY"
    assert len(status["message"]) > 0
    assert status["deployments"]["lowercase_deployment"]["status"] == "UNHEALTHY"
    assert len(status["deployments"]["lowercase_deployment"]["message"]) > 0

    # Wait for the RUNNING status for 30 seconds (deployment should restart and be healthy again)
    for _ in range(30):
        status = get_status(aana_app)
        if status["status"] == "RUNNING":
            break
        await asyncio.sleep(1)
    assert status["status"] == "RUNNING"
    assert len(status["message"]) == 0
    assert status["deployments"]["lowercase_deployment"]["status"] == "RUNNING"
    assert len(status["deployments"]["lowercase_deployment"]["message"]) == 0



# ruff: noqa: S101
import json
import re

import pytest
from pydantic import BaseModel, ValidationError

from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.sampling import SamplingParams
from aana.core.models.types import Dtype
from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.vllm_deployment import VLLMConfig, VLLMDeployment

deployments = [
    (
        "phi3_mini_4k_instruct_vllm_deployment",
        VLLMDeployment.options(
            num_replicas=1,
            max_ongoing_requests=1000,
            ray_actor_options={"num_gpus": 0.5},
            user_config=VLLMConfig(
                model_id="microsoft/Phi-3-mini-4k-instruct",
                dtype=Dtype.FLOAT16,
                gpu_memory_reserved=10000,
                enforce_eager=True,
                default_sampling_params=SamplingParams(
                    temperature=0.0,
                    top_p=1.0,
                    top_k=-1,
                    max_tokens=1024,
                ),
                engine_args={
                    "trust_remote_code": True,
                },
            ).model_dump(mode="json"),
        ),
    ),
]


class CityDescription(BaseModel):
    """A test model for city descriptions."""

    city: str
    country: str
    description: str


@pytest.mark.parametrize(
    "setup_deployment",
    deployments,
    indirect=["setup_deployment"],
)
class TestStructuredGeneration:
    """Test schema and regex constrained generation with vLLM deployment."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("query", ["Tell me about Vienna.", "Describe Berlin."])
    async def test_chat_with_json_schema(self, setup_deployment, query):
        """Test chat method with JSON schema."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        dialog = ChatDialog(
            messages=[
                ChatMessage(role="user", content=query),
            ]
        )

        # Define a JSON schema for CityDescription
        schema = json.dumps(CityDescription.model_json_schema(), indent=2)

        # Test chat method with JSON schema validation
        output = await handle.chat(
            dialog=dialog,
            sampling_params=SamplingParams(
                json_schema=schema, temperature=0.0, max_tokens=512
            ),
        )

        response_message = output["message"]
        assert response_message.role == "assistant"
        text = response_message.content
        print(text)

        # Validate JSON response against the schema
        try:
            CityDescription.model_validate_json(text)
        except ValidationError as e:
            pytest.fail(f"Response does not match schema: {e}")

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "query",
        [
            "Tell me about Paris and New York. Return a list of dictionaries.",
            "Describe Tokyo and London. Return a list of dictionaries.",
        ],
    )
    async def test_chat_with_list_schema(self, setup_deployment, query):
        """Test chat method with list of city descriptions schema."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        dialog = ChatDialog(
            messages=[
                ChatMessage(
                    role="user",
                    content=query,
                ),
            ]
        )

        # Define a JSON schema for a list of CityDescription
        from pydantic import RootModel

        CityDescriptionList = RootModel[list[CityDescription]]
        schema = json.dumps(CityDescriptionList.model_json_schema(), indent=2)

        # Test chat method with JSON schema validation
        output = await handle.chat(
            dialog=dialog,
            sampling_params=SamplingParams(
                json_schema=schema, temperature=0.0, max_tokens=512
            ),
        )

        response_message = output["message"]
        assert response_message.role == "assistant"
        text = response_message.content
        print(text)

        # Validate JSON response against the schema
        try:
            CityDescriptionList.model_validate_json(text)
        except ValidationError as e:
            pytest.fail(f"Response does not match schema: {e}")

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "query, regex_pattern",
        [
            (
                "What is Pi? Give me the first 15 digits. Only return the number.",
                "(-)?(0|[1-9][0-9]*)(\\.[0-9]+)?([eE][+-][0-9]+)?",
            )
        ],
    )
    async def test_chat_with_regex(self, setup_deployment, query, regex_pattern):
        """Test chat method with regex."""
        deployment_name, handle_name, _ = setup_deployment

        handle = await AanaDeploymentHandle.create(handle_name)

        dialog = ChatDialog(
            messages=[
                ChatMessage(
                    role="user",
                    content=query,
                ),
            ]
        )

        # Test chat method with regex validation
        output = await handle.chat(
            dialog=dialog,
            sampling_params=SamplingParams(
                regex_string=regex_pattern, temperature=0.0, max_tokens=32
            ),
        )

        response_message = output["message"]
        assert response_message.role == "assistant"
        text = response_message.content
        print(text)

        # Validate response against the regex pattern
        if not re.fullmatch(regex_pattern, text):
            pytest.fail(f"Response does not match regex pattern: {text}")



from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field

from aana.core.models.time import TimeInterval

__all__ = [
    "PyannoteSpeakerDiarizationParams",
    "SpeakerDiarizationSegment",
    "SpeakerDiarizationSegments",
]


class PyannoteSpeakerDiarizationParams(BaseModel):
    """A model for the pyannote Speaker Diarization model parameters.

    Attributes:
        min_speakers (int | None): The minimum number of speakers present in the audio.
        max_speakers (int | None): The maximum number of speakers present in the audio.
    """

    min_speakers: int | None = Field(
        default=None,
        description="The minimum number of speakers present in the audio.",
    )

    max_speakers: int | None = Field(
        default=None,
        description="The maximum number of speakers present in the audio.",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Parameters for the pyannote speaker diarization model.",
        }
    )


class SpeakerDiarizationSegment(BaseModel):
    """Pydantic schema for Segment from Speaker Diarization model.

    Attributes:
        time_interval (TimeInterval): The start and end time of the segment
        speaker (str): speaker assignment of the model in the format "SPEAKER_XX"
    """

    time_interval: TimeInterval = Field(description="Time interval of the segment")
    speaker: str = Field(description="speaker assignment from the model")

    def to_dict(self) -> dict:
        """Generate dictionary with start, end and speaker keys from SpeakerDiarizationSegment.

        Returns:
            dict: Dictionary with start, end and speaker keys
        """
        return {
            "start": self.time_interval.start,
            "end": self.time_interval.end,
            "speaker": self.speaker,
        }

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Speaker Diarization Segment",
        }
    )


SpeakerDiarizationSegments = Annotated[
    list[SpeakerDiarizationSegment],
    Field(description="List of Speaker Diarization segments", default_factory=list),
]
"""
List of SpeakerDiarizationSegment objects.
"""



from pathlib import Path

import numpy as np


class AbstractAudioLibrary:
    """Abstract class for audio libraries."""

    @classmethod
    def read_file(cls, path: Path) -> np.ndarray:
        """Read an audio file from path and return as numpy audio array.

        Args:
            path (Path): The path of the file to read.

        Returns:
            np.ndarray: The file as a numpy array.
        """
        raise NotImplementedError

    @classmethod
    def read_from_bytes(cls, content: bytes) -> np.ndarray:
        """Read bytes using the image library.

        Args:
            content (bytes): The content of the file to read.

        Returns:
            np.ndarray: The file as a numpy array.
        """
        raise NotImplementedError

    @classmethod
    def write_file(cls, path: Path, audio: np.ndarray):
        """Write a file using the audio library.

        Args:
            path (Path): The path of the file to write.
            audio (np.ndarray): The audio to write.
        """
        raise NotImplementedError

    @classmethod
    def write_to_bytes(cls, audio: np.ndarray) -> bytes:
        """Write bytes using the audio library.

        Args:
            audio (np.ndarray): The audio to write.

        Returns:
            bytes: The audio as bytes.
        """
        raise NotImplementedError

    @classmethod
    def write_audio_bytes(cls, path: Path, audio: bytes, sample_rate: int = 16000):
        """Write a file to wav from the normalized audio bytes.

        Args:
            path (Path): The path of the file to write.
            audio (bytes): The audio to in 16-bit PCM byte write.
            sample_rate (int): The sample rate of the audio.
        """
        raise NotImplementedError



# ruff: noqa: S101, S113
from collections.abc import AsyncGenerator

import pytest
import requests
from openai import NotFoundError, OpenAI
from ray import serve

from aana.core.models.chat import ChatDialog, ChatMessage
from aana.core.models.sampling import SamplingParams
from aana.deployments.base_text_generation_deployment import (
    BaseTextGenerationDeployment,
    ChatOutput,
    LLMOutput,
)


@serve.deployment
class LowercaseLLM(BaseTextGenerationDeployment):
    """Ray deployment that returns the lowercase version of a text structured as an LLM."""

    async def generate_stream(
        self, prompt: str, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Generate text stream.

        Args:
            prompt (str): The prompt.
            sampling_params (SamplingParams): The sampling parameters.

        Yields:
            LLMOutput: The generated text.
        """
        for char in prompt:
            yield LLMOutput(text=char.lower())

    async def chat(
        self, dialog: ChatDialog, sampling_params: SamplingParams | None = None
    ) -> ChatOutput:
        """Dummy chat method."""
        text = dialog.messages[-1].content
        return ChatOutput(message=ChatMessage(content=text.lower(), role="assistant"))

    async def chat_stream(
        self, dialog: ChatDialog, sampling_params: SamplingParams | None = None
    ) -> AsyncGenerator[LLMOutput, None]:
        """Dummy chat stream method."""
        text = dialog.messages[-1].content
        for char in text:
            yield LLMOutput(text=char.lower())


deployments = [
    {
        "name": "lowercase_deployment",
        "instance": LowercaseLLM,
    }
]


def test_chat_completion(create_app):
    """Test the chat completion endpoint for OpenAI compatible API."""
    aana_app = create_app(deployments, [])

    port = aana_app.port
    route_prefix = ""

    # Check that the server is ready
    response = requests.get(f"http://localhost:{port}{route_prefix}/api/ready")
    assert response.status_code == 200
    assert response.json() == {"ready": True}

    messages = [
        {"role": "user", "content": "Hello World!"},
    ]
    expected_output = messages[0]["content"].lower()

    client = OpenAI(
        api_key="token",
        base_url=f"http://localhost:{port}",
    )

    # Test chat completion endpoint
    completion = client.chat.completions.create(
        messages=messages,
        model="lowercase_deployment",
    )
    assert completion.choices[0].message.content == expected_output

    # Test chat completion endpoint with stream
    stream = client.chat.completions.create(
        messages=messages,
        model="lowercase_deployment",
        stream=True,
    )
    generated_text = ""
    for chunk in stream:
        generated_text += chunk.choices[0].delta.content or ""
    assert generated_text == expected_output

    # Test chat completion endpoint with non-existent model
    with pytest.raises(NotFoundError) as exc_info:
        completion = client.chat.completions.create(
            messages=messages,
            model="non_existent_model",
        )
    assert (
        exc_info.value.body["message"]
        == "The model `non_existent_model` does not exist."
    )



# ruff: noqa: S101, S108
from pathlib import Path

from aana.configs.settings import Settings


def test_default_tmp_data_dir():
    """Test that the default temporary data directory is set correctly."""
    settings = Settings()
    assert settings.tmp_data_dir == Path("/tmp/aana_data")


def test_custom_tmp_data_dir(monkeypatch):
    """Test that the custom temporary data directory with environment variable is set correctly."""
    test_path = "/tmp/override/path"
    monkeypatch.setenv("TMP_DATA_DIR", test_path)
    settings = Settings()
    assert settings.tmp_data_dir == Path(test_path)



from dataclasses import dataclass
from typing import Any, TypedDict

import torch
from pyannote.audio import Model
from pydantic import BaseModel, ConfigDict, Field
from ray import serve

from aana.core.models.audio import Audio
from aana.core.models.base import pydantic_protected_fields
from aana.core.models.time import TimeInterval
from aana.core.models.vad import VadParams, VadSegment
from aana.deployments.base_deployment import BaseDeployment, exception_handler
from aana.exceptions.runtime import InferenceException
from aana.processors.vad import BinarizeVadScores, VoiceActivitySegmentation
from aana.utils.download import download_model


@dataclass
class SegmentX:
    """The vad segment format with optional speaker.

    Attributes:
        start (float): The start time of the segment.
        end (float): The end time of the segment.
        speaker (str, optional): The speaker of the segment. Defaults to "".
    """

    start: float
    end: float
    speaker: str = ""


class VadOutput(TypedDict):
    """The output of the VAD model.

    Attributes:
        segments (list[VadSegment]): The VAD segments.
    """

    segments: list[VadSegment]


class VadConfig(BaseModel):
    """The configuration for the vad deployment.

    Attributes:
        model (str): The model url.
        onset (float): The onset threshold. Defaults to 0.500.
        offset (float): The offset threshold. Defaults to 0.363.
        min_duration_on (float): The minimum duration for voice activity. Defaults to 0.1.
        min_duration_off (float): The minimum duration for silence. Defaults to 0.1.
        sample_rate (int): The sample rate of the audio. Defaults to 16000.
    """

    model: str = Field(
        description="The VAD model url.",
    )

    # default_parameters for the model on initialization:
    onset: float = Field(
        default=0.500,
        ge=0.0,
        description="Threshold to decide a positive voice activity.",
    )

    offset: float = Field(
        default=0.363,
        ge=0.0,
        description="Threshold to consider as a silence region.",
    )

    min_duration_on: float = Field(
        default=0.1, ge=0.0, description="Minimum voiced duration."
    )

    min_duration_off: float = Field(
        default=0.1, ge=0.0, description="Minimum duration to consider as silence."
    )

    sample_rate: int = Field(default=16000, description="Sample rate of the audio.")

    model_config = ConfigDict(protected_namespaces=(*pydantic_protected_fields,))


@serve.deployment(health_check_timeout_s=180)
class VadDeployment(BaseDeployment):
    """Deployment to serve VAD models."""

    async def apply_config(self, config: dict[str, Any]):
        """Apply the configuration.

        The method is called when the deployment is created or updated.

        It loads the model and instantiate vad_pipeline.

        The configuration should conform to the VadConfig schema.

        """
        config_obj = VadConfig(**config)
        self.hyperparameters = {
            "onset": config_obj.onset,
            "offset": config_obj.offset,
            "min_duration_on": config_obj.min_duration_on,
            "min_duration_off": config_obj.min_duration_off,
        }
        self.sample_rate = config_obj.sample_rate
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

        # for consistency across multiple runs
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(42)

        # model hash for vad_model
        model_hash = str(config_obj.model).split("/")[-2]
        self.filepath = download_model(config_obj.model, model_hash)

        self.vad_model = Model.from_pretrained(self.filepath, use_auth_token=None)
        # vad_pipeline
        self.vad_pipeline = VoiceActivitySegmentation(
            segmentation=self.vad_model, device=torch.device(self.device)
        )
        self.vad_pipeline.instantiate(self.hyperparameters)

    async def __merge(
        self, segments: list[dict], params: VadParams
    ) -> list[VadSegment]:
        """Merge operation to combine vad segments into chunk_size segments.

        Args:
            segments (list[dict]): The list of segments to merge/split.
            params (VadParams): The parameters for the vad model.  #dict[str, Any]

        Returns:
            list[VadSegment]: The list of segments.
        """
        curr_end = 0.0
        merged_segments: list[VadSegment] = []
        seg_idxs: list[tuple[float, float]] = []
        speaker_idxs: list[str] = []

        if not params.chunk_size > 0:
            chunk_error = "Expected positive value."
            raise ValueError(f"{chunk_error}")

        binarize = BinarizeVadScores(
            max_duration=params.chunk_size,
            onset=self.hyperparameters["onset"],
            offset=self.hyperparameters["offset"],
        )
        segments = binarize(segments)

        segments_list = [
            SegmentX(
                max(0.0, speech_turn.start - 0.1), speech_turn.end + 0.1, "UNKNOWN"
            )
            for speech_turn in segments.itersegments()
        ]
        if len(segments_list) == 0:
            # No active speech found in audio.
            return []

        # To make sure the starting point is the start of the segment.
        curr_start = segments_list[0].start

        for seg in segments_list:
            if seg.end - curr_start > params.chunk_size and curr_end - curr_start > 0:
                merged_segments.append(
                    VadSegment(
                        time_interval=TimeInterval(start=curr_start, end=curr_end),
                        segments=seg_idxs,
                    )
                )
                curr_start = seg.start
                seg_idxs = []
                speaker_idxs = []
            curr_end = seg.end
            seg_idxs.append((seg.start, seg.end))
            speaker_idxs.append(seg.speaker)

        merged_segments.append(
            VadSegment(
                time_interval=TimeInterval(start=curr_start, end=curr_end),
                segments=seg_idxs,
            )
        )
        return merged_segments

    async def __inference(self, audio: Audio) -> list[dict]:
        """Perform voice activity detection inference on the Audio with the vad model.

        Args:
            audio (Audio): The audio to perform vad.

        Returns:
            vad_segments (list[dict]): The list of vad segments.

        Raises:
            InferenceException: If the vad inference fails.
        """
        audio_array = audio.get_numpy()
        vad_input = {
            "waveform": torch.from_numpy(audio_array).unsqueeze(0),
            "sample_rate": self.sample_rate,
        }

        try:
            vad_segments = self.vad_pipeline(vad_input)
        except Exception as e:
            raise InferenceException(self.filepath.name) from e

        return vad_segments

    @exception_handler
    async def asr_preprocess_vad(
        self, audio: Audio, params: VadParams | None = None
    ) -> VadOutput:
        """Perform vad inference to get vad segments and further processing (split and merge) to get segments for batched asr inference.

        Args:
            audio (Audio): The audio to perform vad.
            params (VadParams): The parameters for the vad model.

        Returns:
            VadOutput: Output vad segments to the asr model.

        Raises:
            InferenceException: If the vad inference fails.
        """
        if not params:
            params = VadParams()

        intermediate_vad_segments = await self.__inference(audio)
        asr_vad_segments = await self.__merge(
            intermediate_vad_segments,
            params,
        )
        return VadOutput(segments=asr_vad_segments)



from aana.deployments.aana_deployment_handle import AanaDeploymentHandle
from aana.deployments.base_deployment import BaseDeployment
from aana.deployments.base_text_generation_deployment import (
    BaseTextGenerationDeployment,
    ChatOutput,
    LLMBatchOutput,
    LLMOutput,
)
from aana.deployments.haystack_component_deployment import (
    HaystackComponentDeployment,
    HaystackComponentDeploymentConfig,
    RemoteHaystackComponent,
)
from aana.deployments.hf_blip2_deployment import (
    CaptioningBatchOutput,
    CaptioningOutput,
    HFBlip2Config,
    HFBlip2Deployment,
)
from aana.deployments.hf_pipeline_deployment import (
    HfPipelineConfig,
    HfPipelineDeployment,
)
from aana.deployments.hf_text_generation_deployment import (
    HfTextGenerationConfig,
    HfTextGenerationDeployment,
)
from aana.deployments.hqq_text_generation_deployment import (
    HqqBackend,
    HqqTexGenerationConfig,
    HqqTextGenerationDeployment,
)
from aana.deployments.idefics_2_deployment import Idefics2Config, Idefics2Deployment
from aana.deployments.pyannote_speaker_diarization_deployment import (
    PyannoteSpeakerDiarizationConfig,
    PyannoteSpeakerDiarizationDeployment,
    SpeakerDiarizationOutput,
)
from aana.deployments.vad_deployment import (
    VadConfig,
    VadDeployment,
    VadOutput,
)
from aana.deployments.vllm_deployment import VLLMConfig, VLLMDeployment
from aana.deployments.whisper_deployment import (
    WhisperBatchOutput,
    WhisperComputeType,
    WhisperConfig,
    WhisperDeployment,
    WhisperModelSize,
    WhisperOutput,
)

__all__ = [
    "AanaDeploymentHandle",
    "BaseDeployment",
    "BaseTextGenerationDeployment",
    "HfTextGenerationConfig",
    "HfTextGenerationDeployment",
    "HqqTexGenerationConfig",
    "HqqTextGenerationDeployment",
    "HqqBackend",
    "Idefics2Config",
    "Idefics2Deployment",
    "VLLMConfig",
    "VLLMDeployment",
    "WhisperConfig",
    "WhisperDeployment",
    "PyannoteSpeakerDiarizationConfig",
    "PyannoteSpeakerDiarizationDeployment",
    "SpeakerDiarizationOutput",
    "VadConfig",
    "VadDeployment",
    "HfPipelineConfig",
    "HfPipelineDeployment",
    "RemoteHaystackComponent",
    "HaystackComponentDeploymentConfig",
    "HaystackComponentDeployment",
    "HFBlip2Config",
    "HFBlip2Deployment",
    "ChatOutput",
    "LLMBatchOutput",
    "LLMOutput",
    "WhisperOutput",
    "WhisperBatchOutput",
    "WhisperModelSize",
    "WhisperComputeType",
    "VadOutput",
    "CaptioningOutput",
    "CaptioningBatchOutput",
]



import hashlib  # noqa: I001
from dataclasses import dataclass
from pathlib import Path
import numpy as np
from aana.configs.settings import settings
from aana.exceptions.io import AudioReadingException
from aana.integrations.external.av import pyAVWrapper
from aana.integrations.external.decord import is_audio
from aana.core.models.media import Media
from aana.core.libraries.audio import AbstractAudioLibrary
from aana.utils.download import download_file


@dataclass
class Audio(Media):
    """A class representing an audio.

    At least one of `path`, `url`, or `content` must be provided.
    If `save_on_disk` is True, the audio will be saved on disk automatically.

    Attributes:
        path (Path): the path to the audio file
        url (str): the URL of the audio
        content (bytes): the content of the audio in bytes
        media_id (MediaId): the ID of the audio. If not provided, it will be generated automatically.
        title (str): the title of the audio
        description (str): the description of the audio
        audio_dir (Path): the directory to save the audio in
    """

    title: str = ""
    description: str = ""
    audio_dir: Path | None = settings.audio_dir
    numpy: np.ndarray | None = None
    audio_lib: type[AbstractAudioLibrary] = pyAVWrapper

    def _validate(self):
        """Validate the audio.

        Raises:
            ValueError: if none of 'path', 'url', or 'content' is provided
            AudioReadingException: if the audio is not valid
        """
        # validate the parent class
        super()._validate()

        # check that at least one of 'path', 'url' or 'content' is provided
        if not any(
            [
                self.path is not None,
                self.url is not None,
                self.content is not None,
            ]
        ):
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url' or 'content' must be provided."
            )

        # check that the audio is valid
        if self.path and not self.is_audio():
            raise AudioReadingException(audio=self)

    def is_audio(self) -> bool:
        """Checks if it's a valid audio."""
        if not self.path:
            return False
        return is_audio(self.path)

    def save(self):
        """Save the audio on disk.

        If the audio is already available on disk, do nothing.
        If the audio represented as a byte string, save it on disk.
        If the audio is represented as a URL, download it and save it on disk.
        If the audio is represented as a numpy array, convert it to audio wav and save it on disk.

        First check if the audio is already available on disk, then content, then url, then numpy
        to avoid unnecessary operations (e.g. downloading the audio or converting it to wav).

        Raises:
            ValueError: If none of 'path', 'url', 'content' or 'numpy' is provided.
        """
        if self.path:
            return

        audio_dir = settings.audio_dir
        audio_dir.mkdir(parents=True, exist_ok=True)
        file_path = audio_dir / (self.media_id + ".wav")
        if self.content is not None:
            self._save_from_audio_content(file_path)
        elif self.numpy is not None:
            self._save_from_numpy(file_path)
        elif self.url:
            self._save_from_url(file_path)
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', 'content' or 'numpy' must be provided."
            )
        self.path = file_path
        self.is_saved = True

    def _save_from_audio_content(self, file_path: Path):
        """Save the audio from the content.

        Args:
            file_path (Path): the path to save the audio to
        """
        assert self.content is not None  # noqa: S101
        self.audio_lib.write_audio_bytes(file_path, self.content)

    def _save_from_audio_url(self, file_path):
        """Save the audio from the URL.

        Args:
            file_path (Path): the path to save the audio to

        Raises:
            DownloadError: if the audio can't be downloaded
        """
        assert self.url is not None  # noqa: S101
        content: bytes = download_file(self.url)
        self.audio_lib.write_audio_bytes(file_path, content)

    def _save_from_numpy(self, file_path: Path):
        """Save the audio from numpy on disk.

        Args:
            file_path (Path): The path of the file to write.
        """
        assert self.numpy is not None  # noqa: S101 TODO
        self.audio_lib.write_file(file_path, self.numpy)

    def get_numpy(self) -> np.ndarray:
        """Load the audio as a numpy array.

        Returns:
            np.ndarray: The audio as a numpy array.

        Raises:
            ValueError: If none of 'path', 'url', 'content' or 'numpy' is provided.
            AudioReadingException: If there is an error reading the audio.
        """
        if self.numpy is not None:
            return self.numpy
        elif self.path:
            self._load_numpy_from_path()
        elif self.url:
            self._load_numpy_from_url()
        elif self.content:
            self._load_numpy_from_content()
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', 'content' or 'numpy' must be provided."
            )
        assert self.numpy is not None  # noqa: S101
        return self.numpy

    def _load_numpy_from_path(self):
        """Load the audio as a numpy array from a path.

        Raises:
            AudioReadingException: If there is an error reading the audio.
        """
        assert self.path is not None  # noqa: S101
        try:
            self.numpy = self.audio_lib.read_file(self.path)
        except Exception as e:
            raise AudioReadingException(self) from e

    def _load_numpy_from_audio_bytes(self, audio_bytes: bytes):
        """Load the image as a numpy array from image bytes (downloaded from URL or read from file).

        Raises:
            ImageReadingException: If there is an error reading the image.
        """
        try:
            self.numpy = self.audio_lib.read_from_bytes(audio_bytes)
        except Exception as e:
            raise AudioReadingException(self) from e

    def _load_numpy_from_url(self):
        """Load the audio as a numpy array from a URL.

        Raises:
            AudioReadingException: If there is an error reading the audio.
        """
        assert self.url is not None  # noqa: S101
        content: bytes = download_file(self.url)
        self._load_numpy_from_audio_bytes(content)

    def _load_numpy_from_content(self):
        """Load the image as a numpy array from content.

        Raises:
            ImageReadingException: If there is an error reading the image.
        """
        assert self.content is not None  # noqa: S101
        self._load_numpy_from_audio_bytes(self.content)

    def __repr__(self) -> str:
        """Get the representation of the audio.

        Use md5 hash for the content of the audio if it is available.

        Returns:
            str: the representation of the audio
        """
        content_hash = (
            hashlib.md5(self.content, usedforsecurity=False).hexdigest()
            if self.content
            else None
        )
        return (
            f"Audio(path={self.path}, "
            f"url={self.url}, "
            f"content={content_hash}, "
            f"media_id={self.media_id}, "
            f"title={self.title}, "
            f"description={self.description})"
        )



# ruff: noqa: S101

import uuid
from importlib import resources

import pytest

from aana.core.models.video import Video, VideoMetadata
from aana.exceptions.db import MediaIdAlreadyExistsException, NotFoundException
from aana.storage.repository.video import VideoRepository


@pytest.fixture(scope="function")
def dummy_video():
    """Creates a dummy video for testing."""
    media_id = str(uuid.uuid4())
    path = resources.files("aana.tests.files.videos") / "squirrel.mp4"
    video = Video(path=path, media_id=media_id)
    return video


def test_save_video(db_session, dummy_video):
    """Tests saving a video."""
    video_repo = VideoRepository(db_session)
    video_repo.save(dummy_video)

    video_entity = video_repo.read(dummy_video.media_id)
    assert video_entity
    assert video_entity.id == dummy_video.media_id

    # Try to save the same video again
    with pytest.raises(MediaIdAlreadyExistsException):
        video_repo.save(dummy_video)

    video_repo.delete(dummy_video.media_id)
    with pytest.raises(NotFoundException):
        video_repo.read(dummy_video.media_id)


def test_get_metadata(db_session, dummy_video):
    """Tests getting video metadata."""
    video_repo = VideoRepository(db_session)
    video_repo.save(dummy_video)

    metadata = video_repo.get_metadata(dummy_video.media_id)
    assert isinstance(metadata, VideoMetadata)
    assert metadata.title == dummy_video.title
    assert metadata.description == dummy_video.description
    assert metadata.duration == None

    video_repo.delete(dummy_video.media_id)
    with pytest.raises(NotFoundException):
        video_repo.get_metadata(dummy_video.media_id)



from collections.abc import Generator
from pathlib import Path

import decord
import numpy as np
from decord import DECORDError
from typing_extensions import TypedDict

from aana.core.models.image import Image
from aana.core.models.video import Video, VideoParams
from aana.exceptions.io import VideoReadingException

__all__ = [
    "extract_frames",
    "generate_frames",
    "get_video_duration",
    "is_audio",
    "FramesDict",
]


class FramesDict(TypedDict):
    """Represents a set of frames with ids, timestamps and total duration.

    Attributes:
        frames (list[Image]): the extracted frames
        timestamps (list[float]): the timestamps of the extracted frames
        duration (float): the total duration of the video
        frame_ids (list[int]): the ids of the extracted frames
    """

    frames: list[Image]
    timestamps: list[float]
    duration: float
    frame_ids: list[int]


def extract_frames(video: Video, params: VideoParams) -> FramesDict:
    """Extract frames from a video using decord.

    Args:
        video (Video): the video to extract frames from
        params (VideoParams): the parameters of the video extraction

    Returns:
        FramesDict: a dictionary containing the extracted frames, frame_ids, timestamps, and duration
    """
    device = decord.cpu(0)
    num_threads = 1  # TODO: see if we can use more threads

    num_fps: float = params.extract_fps
    try:
        video_reader = decord.VideoReader(
            str(video.path), ctx=device, num_threads=num_threads
        )
    except DECORDError as video_reader_exception:
        try:
            audio_reader = decord.AudioReader(str(video.path), ctx=device)
            return FramesDict(
                frames=[],
                timestamps=[],
                duration=audio_reader.duration(),
                frame_ids=[],
            )
        except DECORDError:
            raise VideoReadingException(video) from video_reader_exception

    video_fps = video_reader.get_avg_fps()
    num_frames = len(video_reader)
    duration = num_frames / video_fps

    if params.fast_mode_enabled:
        indexes = video_reader.get_key_indices()
    else:
        # num_fps can be smaller than 1 (e.g. 0.5 means 1 frame every 2 seconds)
        indexes = np.arange(0, num_frames, int(video_fps / num_fps))
    timestamps = video_reader.get_frame_timestamp(indexes)[:, 0].tolist()

    frames_array = video_reader.get_batch(indexes).asnumpy()
    frames = []
    for frame_id, frame in enumerate(frames_array):
        img = Image(numpy=frame, media_id=f"{video.media_id}_frame_{frame_id}")
        frames.append(img)

    return FramesDict(
        frames=frames,
        timestamps=timestamps,
        duration=duration,
        frame_ids=list(range(len(frames))),
    )


def get_video_duration(video: Video) -> float:
    """Extract video duration using decord.

    Args:
        video (Video): the video to get its duration

    Returns:
        float: duration of the video

    Raises:
        VideoReadingException: if the file is not readable or a valid multimedia file
    """
    device = decord.cpu(0)
    try:
        video_reader = decord.VideoReader(str(video.path), ctx=device, num_threads=1)
    except DECORDError as video_reader_exception:
        try:
            audio_reader = decord.AudioReader(str(video.path), ctx=device)
            return audio_reader.duration()
        except DECORDError:
            raise VideoReadingException(video) from video_reader_exception

    video_fps = video_reader.get_avg_fps()
    num_frames = len(video_reader)
    duration = num_frames / video_fps
    return duration


def generate_frames(
    video: Video, params: VideoParams, batch_size: int = 8
) -> Generator[FramesDict, None, None]:
    """Generate frames from a video using decord.

    Args:
        video (Video): the video to extract frames from
        params (VideoParams): the parameters of the video extraction
        batch_size (int): the number of frames to yield at each iteration

    Yields:
        FramesDict: a dictionary containing the extracted frames, frame ids, timestamps,
                    and duration for each batch
    Raises:
        VideoReadingException: if the file is not readable or a valid multimedia file
    """
    device = decord.cpu(0)
    num_threads = 1  # TODO: see if we can use more threads

    num_fps: float = params.extract_fps
    is_audio_only = False
    try:
        video_reader = decord.VideoReader(
            str(video.path), ctx=device, num_threads=num_threads
        )
    except DECORDError as video_reader_exception:
        try:
            audio_reader = decord.AudioReader(str(video.path), ctx=device)
            is_audio_only = True
            yield FramesDict(
                frames=[],
                timestamps=[],
                duration=audio_reader.duration(),
                frame_ids=[],
            )

        except DECORDError:
            raise VideoReadingException(video) from video_reader_exception

    if is_audio_only:
        return

    video_fps = video_reader.get_avg_fps()
    num_frames = len(video_reader)
    duration = num_frames / video_fps

    if params.fast_mode_enabled:
        indexes = video_reader.get_key_indices()
    else:
        # num_fps can be smaller than 1 (e.g. 0.5 means 1 frame every 2 seconds)
        indexes = np.arange(0, num_frames, int(video_fps / num_fps))
    timestamps = video_reader.get_frame_timestamp(indexes)[:, 0].tolist()

    for i in range(0, len(indexes), batch_size):
        batch = indexes[i : i + batch_size]
        batch_frames_array = video_reader.get_batch(batch).asnumpy()
        batch_frames = []
        for frame_id, frame in enumerate(batch_frames_array):
            img = Image(numpy=frame, media_id=f"{video.media_id}_frame_{frame_id}")
            batch_frames.append(img)

        batch_timestamps = timestamps[i : i + batch_size]
        yield FramesDict(
            frames=batch_frames,
            frame_ids=list(range(len(batch_frames))),
            timestamps=batch_timestamps,
            duration=duration,
        )


def is_audio(path: Path) -> bool:
    """Checks if it's a valid audio."""
    try:
        decord.AudioReader(str(path))
    except DECORDError:
        return False
    return True



from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field

from aana.core.models.time import TimeInterval

__all__ = ["VadParams", "VadSegment", "VadSegments"]


class VadParams(BaseModel):
    """A model for the Voice Activity Detection model parameters.

    Attributes:
        chunk_size (float): The maximum length of each vad output chunk.
        merge_onset (float): Onset to be used for the merging operation.
        merge_offset (float): "Optional offset to be used for the merging operation.
    """

    chunk_size: float = Field(
        default=30, ge=10.0, description="The maximum length of each vad output chunk."
    )

    merge_onset: float = Field(
        default=0.5, ge=0.0, description="Onset to be used for the merging operation."
    )

    merge_offset: float | None = Field(
        default=None,
        description="Optional offset to be used for the merging operation.",
    )

    model_config = ConfigDict(
        json_schema_extra={
            "description": "Parameters for the voice activity detection model.",
        }
    )


class VadSegment(BaseModel):
    """Pydantic schema for Segment from Voice Activity Detection model.

    Attributes:
        time_interval (TimeInterval): The start and end time of the segment
        segments (list[tuple[float, float]]): smaller voiced segments within a merged vad segment
    """

    time_interval: TimeInterval = Field(description="Time interval of the segment")
    segments: list[tuple[float, float]] = Field(
        description="List of voiced segments within a Segment for ASR"
    )

    def to_whisper_dict(self) -> dict:
        """Generate dictionary with start, end and segments keys from VADSegment for faster whisper.

        Returns:
            dict: Dictionary with start, end and segments keys
        """
        return {
            "start": self.time_interval.start,
            "end": self.time_interval.end,
            "segments": self.segments,
        }

    model_config = ConfigDict(
        json_schema_extra={
            "description": "VAD Segment for ASR",
        }
    )


VadSegments = Annotated[
    list[VadSegment], Field(description="List of VAD segments", default_factory=list)
]
"""
List of VadSegment objects.
"""



from os import PathLike

from pydantic_settings import BaseSettings
from sqlalchemy.engine import Engine
from typing_extensions import TypedDict

from aana.storage.op import DbType, create_database_engine


class SQLiteConfig(TypedDict):
    """Config values for SQLite.

    Attributes:
        path (PathLike): The path to the SQLite database file.
    """

    path: PathLike | str


class PostgreSQLConfig(TypedDict):
    """Config values for PostgreSQL.

    Attributes:
        host (str): The host of the PostgreSQL server.
        port (str): The port of the PostgreSQL server.
        user (str): The user to connect to the PostgreSQL server.
        password (str): The password to connect to the PostgreSQL server.
        database (str): The database name.
    """

    host: str
    port: str
    user: str
    password: str
    database: str


class DbSettings(BaseSettings):
    """Database configuration.

    Attributes:
        datastore_type (DbType | str): The type of the datastore. Default is DbType.SQLITE.
        datastore_config (SQLiteConfig | PostgreSQLConfig): The configuration for the datastore.
            Default is SQLiteConfig(path="/var/lib/aana_data").
        pool_size (int): The number of connections to keep in the pool. Default is 5.
        max_overflow (int): The number of connections that can be created when the pool is exhausted.
            Default is 10.
        pool_recycle (int): The number of seconds a connection can be idle in the pool before it is invalidated.
            Default is 3600.
    """

    datastore_type: DbType | str = DbType.SQLITE
    datastore_config: SQLiteConfig | PostgreSQLConfig = SQLiteConfig(
        path="/var/lib/aana_data"
    )
    pool_size: int = 5
    max_overflow: int = 10
    pool_recycle: int = 3600

    _engine: Engine | None = None

    def get_engine(self):
        """Gets engine. Each instance of DbSettings will create a max.of 1 engine."""
        if not self._engine:
            self._engine = create_database_engine(self)
        return self._engine

    def __getstate__(self):
        """Used by pickle to pickle an object."""
        # We need to remove the "engine" property because SqlAlchemy engines
        # are not picklable
        state = self.__dict__.copy()
        state.pop("engine", None)
        return state

    def __setstate__(self, state):
        """Used to restore a runtime object from pickle; the opposite of __getstate__()."""
        # We don't need to do anything special here, since the engine will be recreated
        # if needed.
        self.__dict__.update(state)



import pickle
from typing import Annotated

from pydantic import BeforeValidator, PlainSerializer

__all__ = ["CustomConfig"]

CustomConfig = Annotated[
    dict,
    PlainSerializer(lambda x: pickle.dumps(x).decode("latin1"), return_type=str),
    BeforeValidator(
        lambda x: x if isinstance(x, dict) else pickle.loads(x.encode("latin1"))  # noqa: S301
    ),
]
"""
A custom configuration field that can be used to pass arbitrary configuration to the deployment.

For example, you can define a custom configuration field in a deployment configuration like this:

```python
class HfPipelineConfig(BaseModel):
    model_id: str
    task: str | None = None
    model_kwargs: CustomConfig = {}
    pipeline_kwargs: CustomConfig = {}
    generation_kwargs: CustomConfig = {}
```

Then you can use the custom configuration field to pass a configuration to the deployment:

```python
HfPipelineConfig(
    model_id="Salesforce/blip2-opt-2.7b",
    model_kwargs={
        "quantization_config": BitsAndBytesConfig(
            load_in_8bit=False, load_in_4bit=True
        ),
    },
)
```
"""



"""Set server default for task.completed_at and task.assigned_at to none and add num_retries.

Revision ID: b9860676dd49
Revises: 5ad873484aa3
Create Date: 2024-08-22 07:54:55.921710

"""
from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b9860676dd49"
down_revision: str | None = "5ad873484aa3"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade database to this revision from previous."""
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.alter_column(
            "completed_at",
            server_default=None,
        )
        batch_op.alter_column(
            "assigned_at",
            server_default=None,
        )
        batch_op.add_column(
            sa.Column(
                "num_retries",
                sa.Integer(),
                nullable=False,
                comment="Number of retries",
                server_default=sa.text("0"),
            )
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database from this revision to previous."""
    with op.batch_alter_table("tasks", schema=None) as batch_op:
        batch_op.drop_column("num_retries")

    # ### end Alembic commands ###



from typing import Any

import orjson
from fastapi.responses import JSONResponse

from aana.utils.json import jsonify


class AanaJSONResponse(JSONResponse):
    """Response class that uses orjson to serialize data.

    It has additional support for numpy arrays.
    """

    media_type = "application/json"
    option = None

    def __init__(self, option: int | None = orjson.OPT_SERIALIZE_NUMPY, **kwargs):
        """Initialize the response class with the orjson option."""
        self.option = option
        super().__init__(**kwargs)

    def render(self, content: Any) -> bytes:
        """Override the render method to use orjson.dumps instead of json.dumps."""
        return jsonify(content, option=self.option, as_bytes=True)



from typing import Any


class BaseException(Exception):  # noqa: A001
    """Base class for SDK exceptions."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialise Exception."""
        super().__init__()
        self.extra = kwargs

    def __str__(self) -> str:
        """Return a string representation of the exception.

        String is defined as follows:
        ```
        <class_name>(extra_key1=extra_value1, extra_key2=extra_value2, ...)
        ```
        """
        class_name = self.__class__.__name__
        extra_str_list = []
        for key, value in self.extra.items():
            extra_str_list.append(f"{key}={value}")
        extra_str = ", ".join(extra_str_list)
        return f"{class_name}({extra_str})"

    def get_data(self) -> dict[str, Any]:
        """Get the data to be returned to the client.

        Returns:
            Dict[str, Any]: data to be returned to the client
        """
        data = self.extra.copy()
        return data

    def add_extra(self, data: dict[str, Any]) -> None:
        """Add extra data to the exception.

        This data will be returned to the user as part of the response.

        How to use: in the exception handler, add the extra data to the exception and raise it again.

        Example:
            ```
            try:
                ...
            except BaseException as e:
                e.add_extra({'extra_key': 'extra_value'})
                raise e
            ```

        Args:
            data (dict[str, Any]): dictionary containing the extra data
        """
        self.extra.update(data)



# ruff: noqa: S101

import pytest
from pydantic import BaseModel, ValidationError

from aana.core.models.media import MediaId


def test_media_id_creation():
    """Test that a media id can be created."""

    class TestModel(BaseModel):
        media_id: MediaId

    media_id = TestModel(media_id="foo").media_id
    assert media_id == "foo"

    # Validation only happens when the model is created
    # because MediaId is just an annotated string
    with pytest.raises(ValueError):
        TestModel().media_id  # noqa: B018

    with pytest.raises(ValidationError):
        TestModel(media_id="").media_id  # noqa: B018

    # MediaId is a string with a maximum length of 36
    with pytest.raises(ValidationError):
        TestModel(media_id="a" * 37).media_id  # noqa: B018



import hashlib
import uuid
from dataclasses import dataclass, field
from pathlib import Path
from typing import Annotated, Any

from pydantic import Field, ValidationInfo, ValidatorFunctionWrapHandler
from pydantic.functional_validators import WrapValidator

from aana.utils.download import download_file


def verify_media_id_must_not_be_empty(
    v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
) -> str:
    """Validates that the media_id is not an empty string."""
    assert v != "", "media_id cannot be an empty string"  # noqa: S101
    return v


def verify_media_id_length(
    v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
) -> str:
    """Validates that the media_id is maximum 36 characters long."""
    assert len(v) <= 36, "media_id must be at most 36 characters long"  # noqa: S101
    return v


def verify_media_id(
    v: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
) -> str:
    """Validates the media_id."""
    v = verify_media_id_must_not_be_empty(v, handler, info)
    v = verify_media_id_length(v, handler, info)
    return v


MediaId = Annotated[
    str,
    Field(description="The media ID.", max_length=36),
    WrapValidator(verify_media_id),
]
"""
The media ID (str, max length 36 characters).
"""


@dataclass
class Media:
    """A base class representing a media file.

    It is used to represent images, videos, and audio files.

    At least one of `path`, `url`, or `content` must be provided.
    If `save_on_disk` is True, the media will be saved on disk automatically.

    Attributes:
        path (Path): the path to the media file
        url (str): the URL of the media
        content (bytes): the content of the media in bytes
        media_id (MediaId): the ID of the media. If not provided, it will be generated automatically.
    """

    path: Path | None = None
    url: str | None = None
    content: bytes | None = None
    media_id: MediaId = field(default_factory=lambda: str(uuid.uuid4()))
    save_on_disk: bool = True
    is_saved: bool = False
    media_dir: Path | None = None

    def _validate(self):
        """Validate the media."""
        # check that path is a Path object
        if self.path and not isinstance(self.path, Path):
            raise ValueError("'path' must be a Path object.")  # noqa: TRY003

        # check if path exists if provided
        if self.path and not self.path.exists():
            raise FileNotFoundError(f"File '{self.path}' does not exist.")  # noqa: TRY003

    def __post_init__(self):
        """Post-initialization.

        Perform checks and save the media on disk if needed.
        """
        self._validate()

        if self.save_on_disk:
            self.save()

    def save(self):
        """Save the media on disk.

        If the media is already available on disk, do nothing.
        If the media represented as a byte string, save it on disk
        If the media is represented as a URL, download it and save it on disk.

        Raises:
            ValueError: if at least one of 'path', 'url', or 'content' is not provided
        """
        if self.path:
            return

        if self.media_dir is None:
            raise ValueError(  # noqa: TRY003
                "The 'media_dir' isn't defined for this media type."
            )
        self.media_dir.mkdir(parents=True, exist_ok=True)
        file_path = self.media_dir / (self.media_id + ".mp4")

        if self.content:
            self._save_from_content(file_path)
        elif self.url:
            self._save_from_url(file_path)
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', or 'content' must be provided."
            )
        self.is_saved = True

    def _save_from_bytes(self, file_path: Path, content: bytes):
        """Save the media from bytes.

        Args:
            file_path (Path): the path to save the media to
            content (bytes): the content of the media
        """
        file_path.write_bytes(content)
        self.path = file_path

    def _save_from_content(self, file_path: Path):
        """Save the media from the content.

        Args:
            file_path (Path): the path to save the media to
        """
        assert self.content is not None  # noqa: S101
        self._save_from_bytes(file_path, self.content)

    def _save_from_url(self, file_path):
        """Save the media from the URL.

        Args:
            file_path (Path): the path to save the media to

        Raises:
            DownloadError: if the media can't be downloaded
        """
        assert self.url is not None  # noqa: S101
        content: bytes = download_file(self.url)
        self._save_from_bytes(file_path, content)

    def get_content(self) -> bytes:
        """Get the content of the media as bytes.

        Returns:
            bytes: the content of the media

        Raises:
            ValueError: if at least one of 'path', 'url', or 'content' is not provided
        """
        if self.content:
            return self.content
        elif self.path:
            self._load_content_from_path()
        elif self.url:
            self._load_content_from_url()
        else:
            raise ValueError(  # noqa: TRY003
                "At least one of 'path', 'url', or 'content' must be provided."
            )
        assert self.content is not None  # noqa: S101
        return self.content

    def _load_content_from_path(self):
        """Load the content of the media from the path."""
        assert self.path is not None  # noqa: S101
        self.content = self.path.read_bytes()

    def _load_content_from_url(self):
        """Load the content of the media from the URL.

        Raises:
            DownloadError: if the media can't be downloaded
        """
        assert self.url is not None  # noqa: S101
        self.content = download_file(self.url)

    def __repr__(self) -> str:
        """Get the representation of the media.

        Use md5 hash for the content of the media if it is available.

        Returns:
            str: the representation of the media
        """
        content_hash = (
            hashlib.md5(self.content, usedforsecurity=False).hexdigest()
            if self.content
            else None
        )
        return (
            f"Media(path={self.path}, "
            f"url={self.url}, "
            f"content={content_hash}, "
            f"media_id={self.media_id})"
        )

    def __str__(self) -> str:
        """Get the string representation of the media.

        Returns:
            str: the string representation of the media
        """
        return self.__repr__()

    def cleanup(self):
        """Cleanup the media.

        If the media is saved on disk by the class, delete it.
        """
        if self.is_saved and self.path:
            self.path.unlink(missing_ok=True)
